#  Ruddr - Robotic Updater for Dynamic DNS Records
#  Copyright (C) 2023 Dominick C. Pastore
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import ipaddress

import pytest

import doubles
import ruddr
from ruddr import PublishError, FatalPublishError


@pytest.fixture
def mock_zone_splitter(mocker):
    doubles.MockZoneSplitter.clear_domains()
    mocker.patch("ruddr.util.ZoneSplitter", new=doubles.MockZoneSplitter)
    return doubles.MockZoneSplitter


class TestGetZones:
    def test_hardcoded_zones(
        self, empty_addrfile, data_dir, mock_zone_splitter
    ):
        """Test zones not fetched when all hosts have hardcoded zones"""
        updater = doubles.MockTwoWayZoneUpdater(
            'test_updater', empty_addrfile, data_dir,
            fetch_zone_ipv4s_result={
                'example.com': [
                    ('', ipaddress.IPv4Address('1.2.3.4'), 1),
                    ('foo', ipaddress.IPv4Address('1.2.3.4'), 2),
                ],
            },
            put_zone_ipv4s_result={'example.com': None},
        )
        updater.init_hosts_and_zones(
            "example.com/example.com foo.example.com/example.com"
        )
        updater.publish_ipv4(ipaddress.IPv4Address('5.6.7.8'))

        assert updater.get_zones_call_count == 0
        assert mock_zone_splitter.split_domains == []
        assert updater.fetch_zone_ipv4s_calls == ['example.com']
        assert updater.put_zone_ipv4s_calls == [
            ('example.com', {
                '': ([ipaddress.IPv4Address('5.6.7.8')], 1),
                'foo': ([ipaddress.IPv4Address('5.6.7.8')], 2),
            }),
        ]

    def test_some_hardcoded(
        self, empty_addrfile, data_dir, mock_zone_splitter
    ):
        """Test hardcoded zones left untouched when others need to be
        fetched"""
        updater = doubles.MockTwoWayZoneUpdater(
            'test_updater', empty_addrfile, data_dir,
            fetch_zone_ipv4s_result={
                'example.com': [
                    ('', ipaddress.IPv4Address('1.2.3.4'), 1),
                    ('foo', ipaddress.IPv4Address('1.2.3.4'), 2),
                ],
                'bar.example.com': [
                    ('foo', ipaddress.IPv4Address('1.2.3.4'), 3),
                ],
            },
            put_zone_ipv4s_result={'example.com': None,
                                   'bar.example.com': None},
            get_zones_result=['example.com', 'bar.example.com'],
        )
        updater.init_hosts_and_zones(
            "example.com foo.bar.example.com/bar.example.com foo.example.com"
        )
        updater.publish_ipv4(ipaddress.IPv4Address('5.6.7.8'))

        assert updater.get_zones_call_count == 1
        assert mock_zone_splitter.split_domains == []
        assert updater.fetch_zone_ipv4s_calls == [
            'example.com',
            'bar.example.com',
        ]
        assert updater.put_zone_ipv4s_calls == [
            ('example.com', {
                '': ([ipaddress.IPv4Address('5.6.7.8')], 1),
                'foo': ([ipaddress.IPv4Address('5.6.7.8')], 2),
            }),
            ('bar.example.com', {
                'foo': ([ipaddress.IPv4Address('5.6.7.8')], 3),
            }),
        ]

    def test_no_hardcoded(
        self, empty_addrfile, data_dir, mock_zone_splitter
    ):
        """Test zones are fetched when no zones are hardcoded"""
        updater = doubles.MockTwoWayZoneUpdater(
            'test_updater', empty_addrfile, data_dir,
            fetch_zone_ipv4s_result={
                'example.com': [
                    ('', ipaddress.IPv4Address('1.2.3.4'), 1),
                    ('foo', ipaddress.IPv4Address('1.2.3.4'), 2),
                ],
                'example.net': [
                    ('foo.bar', ipaddress.IPv4Address('1.2.3.4'), 3),
                ],
            },
            put_zone_ipv4s_result={'example.com': None,
                                   'example.net': None},
            get_zones_result=['example.com', 'example.net'],
        )
        updater.init_hosts_and_zones(
            "example.com foo.bar.example.net foo.example.com"
        )
        updater.publish_ipv4(ipaddress.IPv4Address('5.6.7.8'))

        assert updater.get_zones_call_count == 1
        assert mock_zone_splitter.split_domains == []
        assert updater.fetch_zone_ipv4s_calls == [
            'example.com',
            'example.net',
        ]
        assert updater.put_zone_ipv4s_calls == [
            ('example.com', {
                '': ([ipaddress.IPv4Address('5.6.7.8')], 1),
                'foo': ([ipaddress.IPv4Address('5.6.7.8')], 2),
            }),
            ('example.net', {
                'foo.bar': ([ipaddress.IPv4Address('5.6.7.8')], 3),
            }),
        ]

    def test_get_zones_not_implemented(
        self, empty_addrfile, data_dir, mock_zone_splitter
    ):
        """Test zones are fetched from Public Suffix List when get_zones is not
        implemented"""
        updater = doubles.MockTwoWayZoneUpdater(
            'test_updater', empty_addrfile, data_dir,
            fetch_zone_ipv4s_result={
                'example.com': [
                    ('', ipaddress.IPv4Address('1.2.3.4'), 1),
                    ('foo', ipaddress.IPv4Address('1.2.3.4'), 2),
                ],
                'example.net': [
                    ('foo.bar', ipaddress.IPv4Address('1.2.3.4'), 3),
                ],
            },
            put_zone_ipv4s_result={'example.com': None,
                                   'example.net': None},
        )
        updater.init_hosts_and_zones(
            "example.com foo.bar.example.net foo.example.com"
        )
        updater.publish_ipv4(ipaddress.IPv4Address('5.6.7.8'))

        assert updater.get_zones_call_count == 1
        assert mock_zone_splitter.split_domains == [
            'example.com',
            'foo.bar.example.net',
            'foo.example.com',
        ]
        assert updater.fetch_zone_ipv4s_calls == [
            'example.com',
            'example.net',
        ]
        assert updater.put_zone_ipv4s_calls == [
            ('example.com', {
                '': ([ipaddress.IPv4Address('5.6.7.8')], 1),
                'foo': ([ipaddress.IPv4Address('5.6.7.8')], 2),
            }),
            ('example.net', {
                'foo.bar': ([ipaddress.IPv4Address('5.6.7.8')], 3),
            }),
        ]

    def test_publish_error(
        self, empty_addrfile, data_dir, mock_zone_splitter
    ):
        """Test publish_ipv4 stops and raises PublishError when get_zones
        raises PublishError"""
        updater = doubles.MockTwoWayZoneUpdater(
            'test_updater', empty_addrfile, data_dir,
            fetch_zone_ipv4s_result={
                'example.com': [
                    ('', ipaddress.IPv4Address('1.2.3.4'), 1),
                    ('foo', ipaddress.IPv4Address('1.2.3.4'), 2),
                ],
                'example.net': [
                    ('foo.bar', ipaddress.IPv4Address('1.2.3.4'), 3),
                ],
            },
            put_zone_ipv4s_result={'example.com': None,
                                   'example.net': None},
            get_zones_result=PublishError,
        )
        updater.init_hosts_and_zones(
            "example.com foo.bar.example.net foo.example.com"
        )
        with pytest.raises(PublishError):
            updater.publish_ipv4(ipaddress.IPv4Address('5.6.7.8'))

        assert updater.get_zones_call_count == 1
        assert mock_zone_splitter.split_domains == []
        assert updater.fetch_zone_ipv4s_calls == []
        assert updater.put_zone_ipv4s_calls == []

    def test_no_matching_zone(
        self, empty_addrfile, data_dir, mock_zone_splitter
    ):
        """Test PublishError is raised when one of the configured hosts does
        not match any zone"""
        updater = doubles.MockTwoWayZoneUpdater(
            'test_updater', empty_addrfile, data_dir,
            fetch_zone_ipv4s_result={
                'example.com': [
                    ('', ipaddress.IPv4Address('1.2.3.4'), 1),
                    ('foo', ipaddress.IPv4Address('1.2.3.4'), 2),
                ],
                'example.net': [
                    ('foo.bar', ipaddress.IPv4Address('1.2.3.4'), 3),
                ],
            },
            put_zone_ipv4s_result={'example.com': None,
                                   'example.net': None},
            get_zones_result=['example.com', 'example.net'],
        )
        updater.init_hosts_and_zones(
            "example.com foo.bar.example.net foo.example.com bar.example.org"
        )
        with pytest.raises(PublishError):
            updater.publish_ipv4(ipaddress.IPv4Address('5.6.7.8'))

        assert updater.get_zones_call_count == 1
        assert mock_zone_splitter.split_domains == []
        assert updater.fetch_zone_ipv4s_calls == [
            'example.com',
            'example.net',
        ]
        assert updater.put_zone_ipv4s_calls == [
            ('example.com', {
                '': ([ipaddress.IPv4Address('5.6.7.8')], 1),
                'foo': ([ipaddress.IPv4Address('5.6.7.8')], 2),
            }),
            ('example.net', {
                'foo.bar': ([ipaddress.IPv4Address('5.6.7.8')], 3),
            }),
        ]

    def test_extra_zone(
        self, empty_addrfile, data_dir, mock_zone_splitter
    ):
        """Test extra zones that don't match any host are present from
        get_zones and their records are not fetched"""
        updater = doubles.MockTwoWayZoneUpdater(
            'test_updater', empty_addrfile, data_dir,
            fetch_zone_ipv4s_result={
                'example.com': [
                    ('', ipaddress.IPv4Address('1.2.3.4'), 1),
                    ('foo', ipaddress.IPv4Address('1.2.3.4'), 2),
                ],
                'example.net': [
                    ('foo.bar', ipaddress.IPv4Address('1.2.3.4'), 3),
                ],
            },
            put_zone_ipv4s_result={'example.com': None,
                                   'example.net': None},
            get_zones_result=['example.com', 'example.net', 'example.org'],
        )
        updater.init_hosts_and_zones(
            "example.com foo.bar.example.net foo.example.com"
        )
        updater.publish_ipv4(ipaddress.IPv4Address('5.6.7.8'))

        assert updater.get_zones_call_count == 1
        assert mock_zone_splitter.split_domains == []
        assert updater.fetch_zone_ipv4s_calls == [
            'example.com',
            'example.net',
        ]
        assert updater.put_zone_ipv4s_calls == [
            ('example.com', {
                '': ([ipaddress.IPv4Address('5.6.7.8')], 1),
                'foo': ([ipaddress.IPv4Address('5.6.7.8')], 2),
            }),
            ('example.net', {
                'foo.bar': ([ipaddress.IPv4Address('5.6.7.8')], 3),
            }),
        ]

    @pytest.mark.parametrize(('host', 'zones', 'subdomain', 'zone'), [
        ('bar.example.com',
         ['example.com', 'bar.example.com', 'example.net'],
         '',
         'bar.example.com'),
        ('bar.example.com',
         ['bar.example.com', 'example.com', 'example.net'],
         '',
         'bar.example.com'),
        ('foo.bar.example.com',
         ['example.com', 'bar.example.com', 'example.net'],
         'foo',
         'bar.example.com'),
        ('foo.bar.example.com',
         ['bar.example.com', 'example.com', 'example.net'],
         'foo',
         'bar.example.com'),
    ])
    def test_longest_zone_matched(
        self, host, zones, subdomain, zone,
        empty_addrfile, data_dir, mock_zone_splitter,
    ):
        """Test that a host is always matched with the longest zone possible"""
        updater = doubles.MockTwoWayZoneUpdater(
            'test_updater', empty_addrfile, data_dir,
            fetch_zone_ipv4s_result={
                'example.com': [
                    ('', ipaddress.IPv4Address('1.2.3.4'), 1),
                    ('foo', ipaddress.IPv4Address('1.2.3.4'), 2),
                ],
                'example.net': [
                    ('foo.bar', ipaddress.IPv4Address('1.2.3.4'), 3),
                ],
                zone: [
                    (subdomain, ipaddress.IPv4Address('1.2.3.4'), 4),
                ]
            },
            put_zone_ipv4s_result={'example.com': None,
                                   'example.net': None,
                                   zone: None},
            get_zones_result=zones,
        )
        updater.init_hosts_and_zones(
            f"example.com foo.bar.example.net foo.example.com {host}"
        )
        updater.publish_ipv4(ipaddress.IPv4Address('5.6.7.8'))

        assert updater.get_zones_call_count == 1
        assert mock_zone_splitter.split_domains == []
        assert sorted(updater.fetch_zone_ipv4s_calls) == sorted(zones)
        assert sorted(updater.put_zone_ipv4s_calls) == sorted([
            ('example.com', {
                '': ([ipaddress.IPv4Address('5.6.7.8')], 1),
                'foo': ([ipaddress.IPv4Address('5.6.7.8')], 2),
            }),
            ('example.net', {
                'foo.bar': ([ipaddress.IPv4Address('5.6.7.8')], 3),
            }),
            (zone, {
                subdomain: ([ipaddress.IPv4Address('5.6.7.8')], 4),
            }),
        ])


class TestFetchZoneIPv4sImplemented:
    def test_fetch_and_put_zone_preferred(self, empty_addrfile, data_dir):
        """Test that fetch_zone_ipv4s and put_zone_ipv4s are preferred over
        fetch_subdomain_ipv4s and put_subdomain_ipv4 when all are
        implemented"""
        updater = doubles.MockTwoWayZoneUpdater(
            'test_updater', empty_addrfile, data_dir,
            fetch_zone_ipv4s_result={
                'example.com': [
                    ('', ipaddress.IPv4Address('1.2.3.4'), 1),
                    ('foo', ipaddress.IPv4Address('1.2.3.4'), 2),
                ],
                'example.net': [
                    ('foo.bar', ipaddress.IPv4Address('1.2.3.4'), 3),
                ],
            },
            fetch_subdomain_ipv4s_result={
                ('', 'example.com'): [
                    (ipaddress.IPv4Address('1.2.3.4'), 1),
                ],
                ('foo', 'example.com'): [
                    (ipaddress.IPv4Address('1.2.3.4'), 2),
                ],
                ('foo.bar', 'example.net'): [
                    (ipaddress.IPv4Address('1.2.3.4'), 3),
                ],
            },
            put_zone_ipv4s_result={'example.com': None,
                                   'example.net': None},
            put_subdomain_ipv4_result={('', 'example.com'): None,
                                       ('foo', 'example.com'): None,
                                       ('foo.bar', 'example.net'): None},
        )
        updater.init_hosts_and_zones(
            "example.com foo.bar.example.net foo.example.com"
        )
        updater.publish_ipv4(ipaddress.IPv4Address('5.6.7.8'))

        assert updater.fetch_zone_ipv4s_calls == [
            'example.com',
            'example.net',
        ]
        assert updater.fetch_subdomain_ipv4s_calls == []
        assert updater.put_zone_ipv4s_calls == [
            ('example.com', {
                '': ([ipaddress.IPv4Address('5.6.7.8')], 1),
                'foo': ([ipaddress.IPv4Address('5.6.7.8')], 2),
            }),
            ('example.net', {
                'foo.bar': ([ipaddress.IPv4Address('5.6.7.8')], 3),
            }),
        ]
        assert updater.put_subdomain_ipv4_calls == []

    def test_extra_records_untouched(self, empty_addrfile, data_dir):
        """Test extra records from fetch_zone_ipv4s are passed to
        put_zone_ipv4s untouched"""
        updater = doubles.MockTwoWayZoneUpdater(
            'test_updater', empty_addrfile, data_dir,
            fetch_zone_ipv4s_result={
                'example.com': [
                    ('', ipaddress.IPv4Address('1.2.3.4'), 1),
                    ('foo', ipaddress.IPv4Address('1.2.3.4'), 2),
                    ('bar', ipaddress.IPv4Address('1.2.3.4'), 4),
                ],
                'example.net': [
                    ('foo.bar', ipaddress.IPv4Address('1.2.3.4'), 3),
                    ('baz', ipaddress.IPv4Address('3.4.5.6'), 5),
                ],
            },
            put_zone_ipv4s_result={'example.com': None,
                                   'example.net': None},
        )
        updater.init_hosts_and_zones(
            "example.com foo.bar.example.net foo.example.com"
        )
        updater.publish_ipv4(ipaddress.IPv4Address('5.6.7.8'))

        assert updater.fetch_zone_ipv4s_calls == [
            'example.com',
            'example.net',
        ]
        assert updater.fetch_subdomain_ipv4s_calls == []
        assert updater.put_zone_ipv4s_calls == [
            ('example.com', {
                '': ([ipaddress.IPv4Address('5.6.7.8')], 1),
                'foo': ([ipaddress.IPv4Address('5.6.7.8')], 2),
                'bar': ([ipaddress.IPv4Address('1.2.3.4')], 4),
            }),
            ('example.net', {
                'foo.bar': ([ipaddress.IPv4Address('5.6.7.8')], 3),
                'baz': ([ipaddress.IPv4Address('3.4.5.6')], 5),
            }),
        ]
        assert updater.put_subdomain_ipv4_calls == []

    def test_missing_record(self, empty_addrfile, data_dir):
        """Test missing record from fetch_zone_ipv4s is PublishError but other
        records still updated"""
        updater = doubles.MockTwoWayZoneUpdater(
            'test_updater', empty_addrfile, data_dir,
            fetch_zone_ipv4s_result={
                'example.com': [
                    ('foo', ipaddress.IPv4Address('1.2.3.4'), 2),
                ],
                'example.net': [
                    ('foo.bar', ipaddress.IPv4Address('1.2.3.4'), 3),
                ],
            },
            put_zone_ipv4s_result={'example.com': None,
                                   'example.net': None},
        )
        updater.init_hosts_and_zones(
            "example.com foo.bar.example.net foo.example.com"
        )
        with pytest.raises(PublishError):
            updater.publish_ipv4(ipaddress.IPv4Address('5.6.7.8'))

        assert updater.fetch_zone_ipv4s_calls == [
            'example.com',
            'example.net',
        ]
        assert updater.fetch_subdomain_ipv4s_calls == []
        assert updater.put_zone_ipv4s_calls == [
            ('example.com', {
                'foo': ([ipaddress.IPv4Address('5.6.7.8')], 2),
            }),
            ('example.net', {
                'foo.bar': ([ipaddress.IPv4Address('5.6.7.8')], 3),
            }),
        ]
        assert updater.put_subdomain_ipv4_calls == []

    def test_multiple_records_to_single(self, empty_addrfile, data_dir):
        """Test when multiple records are set for a host, they are replaced
        with a single record, but multiple records on irrelevant hosts are left
        alone."""
        updater = doubles.MockTwoWayZoneUpdater(
            'test_updater', empty_addrfile, data_dir,
            fetch_zone_ipv4s_result={
                'example.com': [
                    # Different TTL within an RR set is not valid, but some
                    # providers may support it anyway. We take the minimum.
                    ('', ipaddress.IPv4Address('1.2.3.4'), 1),
                    ('', ipaddress.IPv4Address('2.3.4.5'), 11),
                    ('foo', ipaddress.IPv4Address('1.2.3.4'), 2),
                    ('foo', ipaddress.IPv4Address('2.3.4.5'), 22),
                    ('bar', ipaddress.IPv4Address('1.2.3.4'), 4),
                    ('bar', ipaddress.IPv4Address('2.3.4.5'), 44),
                ],
                'example.net': [
                    ('foo.bar', ipaddress.IPv4Address('1.2.3.4'), 3),
                    ('foo.bar', ipaddress.IPv4Address('2.3.4.5'), 33),
                    ('baz', ipaddress.IPv4Address('3.4.5.6'), 5),
                    ('baz', ipaddress.IPv4Address('4.5.6.7'), 55),
                ],
            },
            put_zone_ipv4s_result={'example.com': None,
                                   'example.net': None},
        )
        updater.init_hosts_and_zones(
            "example.com foo.bar.example.net foo.example.com"
        )
        updater.publish_ipv4(ipaddress.IPv4Address('5.6.7.8'))

        assert updater.fetch_zone_ipv4s_calls == [
            'example.com',
            'example.net',
        ]
        assert updater.fetch_subdomain_ipv4s_calls == []
        assert updater.put_zone_ipv4s_calls == [
            ('example.com', {
                '': ([ipaddress.IPv4Address('5.6.7.8')], 1),
                'foo': ([ipaddress.IPv4Address('5.6.7.8')], 2),
                'bar': ([ipaddress.IPv4Address('1.2.3.4'),
                         ipaddress.IPv4Address('2.3.4.5')], 4),
            }),
            ('example.net', {
                'foo.bar': ([ipaddress.IPv4Address('5.6.7.8')], 3),
                'baz': ([ipaddress.IPv4Address('3.4.5.6'),
                         ipaddress.IPv4Address('4.5.6.7')], 5),
            }),
        ]
        assert updater.put_subdomain_ipv4_calls == []

    @pytest.mark.parametrize(('zone', 'error'), [
        ('example.com', PublishError),
        ('example.com', FatalPublishError),
        ('example.net', PublishError),
        ('example.net', FatalPublishError),
    ])
    def test_put_zone_publish_error(self, empty_addrfile, data_dir,
                                    zone, error):
        """Test other zones are still published after PublishError for one
        zone's put_zone_ipv4s"""
        updater = doubles.MockTwoWayZoneUpdater(
            'test_updater', empty_addrfile, data_dir,
            fetch_zone_ipv4s_result={
                'example.com': [
                    ('', ipaddress.IPv4Address('1.2.3.4'), 1),
                    ('foo', ipaddress.IPv4Address('1.2.3.4'), 2),
                ],
                'example.net': [
                    ('foo.bar', ipaddress.IPv4Address('1.2.3.4'), 3),
                ],
            },
            put_zone_ipv4s_result={'example.com': None,
                                   'example.net': None,
                                   zone: error},
        )
        updater.init_hosts_and_zones(
            "example.com foo.bar.example.net foo.example.com"
        )
        with pytest.raises(error):
            updater.publish_ipv4(ipaddress.IPv4Address('5.6.7.8'))

        assert updater.fetch_zone_ipv4s_calls == [
            'example.com',
            'example.net',
        ]
        assert updater.fetch_subdomain_ipv4s_calls == []
        assert updater.put_zone_ipv4s_calls == [
            ('example.com', {
                '': ([ipaddress.IPv4Address('5.6.7.8')], 1),
                'foo': ([ipaddress.IPv4Address('5.6.7.8')], 2),
            }),
            ('example.net', {
                'foo.bar': ([ipaddress.IPv4Address('5.6.7.8')], 3),
            }),
        ]
        assert updater.put_subdomain_ipv4_calls == []

    @pytest.mark.parametrize(('zone', 'error'), [
        ('example.com', PublishError),
        ('example.com', FatalPublishError),
        ('example.net', PublishError),
        ('example.net', FatalPublishError),
    ])
    def test_fetch_zone_publish_error(self, empty_addrfile, data_dir,
                                      zone, error):
        """Test other zones still updated when fetch_zone_ipv4s raises
        PublishError for one zone"""
        updater = doubles.MockTwoWayZoneUpdater(
            'test_updater', empty_addrfile, data_dir,
            fetch_zone_ipv4s_result={
                'example.com': [
                    ('', ipaddress.IPv4Address('1.2.3.4'), 1),
                    ('foo', ipaddress.IPv4Address('1.2.3.4'), 2),
                ],
                'example.net': [
                    ('foo.bar', ipaddress.IPv4Address('1.2.3.4'), 3),
                ],
                zone: error,
            },
            put_zone_ipv4s_result={'example.com': None,
                                   'example.net': None},
        )
        updater.init_hosts_and_zones(
            "example.com foo.bar.example.net foo.example.com"
        )
        with pytest.raises(error):
            updater.publish_ipv4(ipaddress.IPv4Address('5.6.7.8'))

        assert updater.fetch_zone_ipv4s_calls == [
            'example.com',
            'example.net',
        ]
        assert updater.fetch_subdomain_ipv4s_calls == []
        calls = [
            ('example.com', {
                '': ([ipaddress.IPv4Address('5.6.7.8')], 1),
                'foo': ([ipaddress.IPv4Address('5.6.7.8')], 2),
            }),
            ('example.net', {
                'foo.bar': ([ipaddress.IPv4Address('5.6.7.8')], 3),
            }),
        ]
        assert updater.put_zone_ipv4s_calls == [
            c for c in calls if c[0] != zone
        ]
        assert updater.put_subdomain_ipv4_calls == []

    def test_put_zone_not_implemented(self, empty_addrfile, data_dir):
        """Test that put_subdomain_ipv4 is used when fetch_zone_ipv4s is
        implemented but put_zone_ipv4 is not"""
        updater = doubles.MockTwoWayZoneUpdater(
            'test_updater', empty_addrfile, data_dir,
            fetch_zone_ipv4s_result={
                'example.com': [
                    ('', ipaddress.IPv4Address('1.2.3.4'), 1),
                    ('foo', ipaddress.IPv4Address('1.2.3.4'), 2),
                ],
                'example.net': [
                    ('foo.bar', ipaddress.IPv4Address('1.2.3.4'), 3),
                ],
            },
            put_subdomain_ipv4_result={('', 'example.com'): None,
                                       ('foo', 'example.com'): None,
                                       ('foo.bar', 'example.net'): None},
        )
        updater.init_hosts_and_zones(
            "example.com foo.bar.example.net foo.example.com"
        )
        updater.publish_ipv4(ipaddress.IPv4Address('5.6.7.8'))

        assert updater.fetch_zone_ipv4s_calls == [
            'example.com',
            'example.net',
        ]
        assert updater.fetch_subdomain_ipv4s_calls == []
        assert updater.put_zone_ipv4s_calls == [
            ('example.com', {
                '': ([ipaddress.IPv4Address('5.6.7.8')], 1),
                'foo': ([ipaddress.IPv4Address('5.6.7.8')], 2),
            }),
            ('example.net', {
                'foo.bar': ([ipaddress.IPv4Address('5.6.7.8')], 3),
            }),
        ]
        assert updater.put_subdomain_ipv4_calls == [
            ('', 'example.com', ipaddress.IPv4Address('5.6.7.8'), 1),
            ('foo', 'example.com', ipaddress.IPv4Address('5.6.7.8'), 2),
            ('foo.bar', 'example.net', ipaddress.IPv4Address('5.6.7.8'), 3),
        ]

    def test_put_zone_not_implemented_extra_records(self, empty_addrfile,
                                                    data_dir):
        """Test that put_subdomain_ipv4 is not called for extra records
        returned by fetch_zone_ipv4s"""
        updater = doubles.MockTwoWayZoneUpdater(
            'test_updater', empty_addrfile, data_dir,
            fetch_zone_ipv4s_result={
                'example.com': [
                    ('', ipaddress.IPv4Address('1.2.3.4'), 1),
                    ('foo', ipaddress.IPv4Address('1.2.3.4'), 2),
                    ('bar', ipaddress.IPv4Address('1.2.3.4'), 4),
                ],
                'example.net': [
                    ('foo.bar', ipaddress.IPv4Address('1.2.3.4'), 3),
                    ('baz', ipaddress.IPv4Address('3.4.5.6'), 5),
                ],
            },
            put_subdomain_ipv4_result={('', 'example.com'): None,
                                       ('foo', 'example.com'): None,
                                       ('foo.bar', 'example.net'): None},
        )
        updater.init_hosts_and_zones(
            "example.com foo.bar.example.net foo.example.com"
        )
        updater.publish_ipv4(ipaddress.IPv4Address('5.6.7.8'))

        assert updater.fetch_zone_ipv4s_calls == [
            'example.com',
            'example.net',
        ]
        assert updater.fetch_subdomain_ipv4s_calls == []
        assert updater.put_zone_ipv4s_calls == [
            ('example.com', {
                '': ([ipaddress.IPv4Address('5.6.7.8')], 1),
                'foo': ([ipaddress.IPv4Address('5.6.7.8')], 2),
                'bar': ([ipaddress.IPv4Address('1.2.3.4')], 4),
            }),
            ('example.net', {
                'foo.bar': ([ipaddress.IPv4Address('5.6.7.8')], 3),
                'baz': ([ipaddress.IPv4Address('3.4.5.6')], 5),
            }),
        ]
        assert updater.put_subdomain_ipv4_calls == [
            ('', 'example.com', ipaddress.IPv4Address('5.6.7.8'), 1),
            ('foo', 'example.com', ipaddress.IPv4Address('5.6.7.8'), 2),
            ('foo.bar', 'example.net', ipaddress.IPv4Address('5.6.7.8'), 3),
        ]

    def test_put_zone_not_implemented_missing_record(self, empty_addrfile,
                                                     data_dir):
        """Test missing record from fetch_zone_ipv4s is PublishError but other
        records still updated with put_subdomain_ipv4 when put_zone_ipv4s not
        implemented"""
        updater = doubles.MockTwoWayZoneUpdater(
            'test_updater', empty_addrfile, data_dir,
            fetch_zone_ipv4s_result={
                'example.com': [
                    ('foo', ipaddress.IPv4Address('1.2.3.4'), 2),
                ],
                'example.net': [
                    ('foo.bar', ipaddress.IPv4Address('1.2.3.4'), 3),
                ],
            },
            put_subdomain_ipv4_result={('foo', 'example.com'): None,
                                       ('foo.bar', 'example.net'): None},
        )
        updater.init_hosts_and_zones(
            "example.com foo.bar.example.net foo.example.com"
        )
        with pytest.raises(PublishError):
            updater.publish_ipv4(ipaddress.IPv4Address('5.6.7.8'))

        assert updater.fetch_zone_ipv4s_calls == [
            'example.com',
            'example.net',
        ]
        assert updater.fetch_subdomain_ipv4s_calls == []
        assert updater.put_zone_ipv4s_calls == [
            ('example.com', {
                'foo': ([ipaddress.IPv4Address('5.6.7.8')], 2),
            }),
            ('example.net', {
                'foo.bar': ([ipaddress.IPv4Address('5.6.7.8')], 3),
            }),
        ]
        assert updater.put_subdomain_ipv4_calls == [
            ('foo', 'example.com', ipaddress.IPv4Address('5.6.7.8'), 2),
            ('foo.bar', 'example.net', ipaddress.IPv4Address('5.6.7.8'), 3),
        ]

    def test_put_zone_not_implemented_multiple_records_to_single(
        self, empty_addrfile, data_dir
    ):
        """Test when multiple records are set for a host and put_zone_ipv4s is
        not implemented, put_subdomain_ipv4 is called exactly once for each
        relevant host and not at all for irrelevant hosts"""
        updater = doubles.MockTwoWayZoneUpdater(
            'test_updater', empty_addrfile, data_dir,
            fetch_zone_ipv4s_result={
                'example.com': [
                    # Different TTL within an RR set is not valid, but some
                    # providers may support it anyway. We take the minimum.
                    ('', ipaddress.IPv4Address('1.2.3.4'), 1),
                    ('', ipaddress.IPv4Address('2.3.4.5'), 11),
                    ('foo', ipaddress.IPv4Address('1.2.3.4'), 2),
                    ('foo', ipaddress.IPv4Address('2.3.4.5'), 22),
                    ('bar', ipaddress.IPv4Address('1.2.3.4'), 4),
                    ('bar', ipaddress.IPv4Address('2.3.4.5'), 44),
                ],
                'example.net': [
                    ('foo.bar', ipaddress.IPv4Address('1.2.3.4'), 3),
                    ('foo.bar', ipaddress.IPv4Address('2.3.4.5'), 33),
                    ('baz', ipaddress.IPv4Address('3.4.5.6'), 5),
                    ('baz', ipaddress.IPv4Address('4.5.6.7'), 55),
                ],
            },
            put_subdomain_ipv4_result={('', 'example.com'): None,
                                       ('foo', 'example.com'): None,
                                       ('foo.bar', 'example.net'): None},
        )
        updater.init_hosts_and_zones(
            "example.com foo.bar.example.net foo.example.com"
        )
        updater.publish_ipv4(ipaddress.IPv4Address('5.6.7.8'))

        assert updater.fetch_zone_ipv4s_calls == [
            'example.com',
            'example.net',
        ]
        assert updater.fetch_subdomain_ipv4s_calls == []
        assert updater.put_zone_ipv4s_calls == [
            ('example.com', {
                '': ([ipaddress.IPv4Address('5.6.7.8')], 1),
                'foo': ([ipaddress.IPv4Address('5.6.7.8')], 2),
                'bar': ([ipaddress.IPv4Address('1.2.3.4'),
                         ipaddress.IPv4Address('2.3.4.5')], 4),
            }),
            ('example.net', {
                'foo.bar': ([ipaddress.IPv4Address('5.6.7.8')], 3),
                'baz': ([ipaddress.IPv4Address('3.4.5.6'),
                         ipaddress.IPv4Address('4.5.6.7')], 5),
            }),
        ]
        assert updater.put_subdomain_ipv4_calls == [
            ('', 'example.com', ipaddress.IPv4Address('5.6.7.8'), 1),
            ('foo', 'example.com', ipaddress.IPv4Address('5.6.7.8'), 2),
            ('foo.bar', 'example.net', ipaddress.IPv4Address('5.6.7.8'), 3),
        ]

    @pytest.mark.parametrize(('subdomain', 'zone', 'error'), [
        ('', 'example.com', PublishError),
        ('', 'example.com', FatalPublishError),
        ('foo', 'example.com', PublishError),
        ('foo', 'example.com', FatalPublishError),
        ('foo.bar', 'example.net', PublishError),
        ('foo.bar', 'example.net', FatalPublishError),
    ])
    def test_put_zone_not_implemented_put_subdomain_publish_error(
        self, empty_addrfile, data_dir, subdomain, zone, error
    ):
        """Test that remaining subdomains and zones are still published when
        put_subdomain_ipv4 has a PublishError"""
        updater = doubles.MockTwoWayZoneUpdater(
            'test_updater', empty_addrfile, data_dir,
            fetch_zone_ipv4s_result={
                'example.com': [
                    ('', ipaddress.IPv4Address('1.2.3.4'), 1),
                    ('foo', ipaddress.IPv4Address('1.2.3.4'), 2),
                ],
                'example.net': [
                    ('foo.bar', ipaddress.IPv4Address('1.2.3.4'), 3),
                ],
            },
            put_subdomain_ipv4_result={('', 'example.com'): None,
                                       ('foo', 'example.com'): None,
                                       ('foo.bar', 'example.net'): None,
                                       (subdomain, zone): error},
        )
        updater.init_hosts_and_zones(
            "example.com foo.bar.example.net foo.example.com"
        )
        with pytest.raises(error):
            updater.publish_ipv4(ipaddress.IPv4Address('5.6.7.8'))

        assert updater.fetch_zone_ipv4s_calls == [
            'example.com',
            'example.net',
        ]
        assert updater.fetch_subdomain_ipv4s_calls == []
        assert updater.put_zone_ipv4s_calls == [
            ('example.com', {
                '': ([ipaddress.IPv4Address('5.6.7.8')], 1),
                'foo': ([ipaddress.IPv4Address('5.6.7.8')], 2),
            }),
            ('example.net', {
                'foo.bar': ([ipaddress.IPv4Address('5.6.7.8')], 3),
            }),
        ]
        assert updater.put_subdomain_ipv4_calls == [
            ('', 'example.com', ipaddress.IPv4Address('5.6.7.8'), 1),
            ('foo', 'example.com', ipaddress.IPv4Address('5.6.7.8'), 2),
            ('foo.bar', 'example.net', ipaddress.IPv4Address('5.6.7.8'), 3),
        ]

    @pytest.mark.parametrize(('zone', 'error'), [
        ('example.com', PublishError),
        ('example.com', FatalPublishError),
        ('example.net', PublishError),
        ('example.net', FatalPublishError),
    ])
    def test_put_zone_not_implemented_fetch_zone_publish_error(
        self, empty_addrfile, data_dir, zone, error
    ):
        """Test other zones still updated when fetch_zone_ipv4s raises
        PublishError for one zone and put_zone_ipv4s not implemented"""
        updater = doubles.MockTwoWayZoneUpdater(
            'test_updater', empty_addrfile, data_dir,
            fetch_zone_ipv4s_result={
                'example.com': [
                    ('', ipaddress.IPv4Address('1.2.3.4'), 1),
                    ('foo', ipaddress.IPv4Address('1.2.3.4'), 2),
                ],
                'example.net': [
                    ('foo.bar', ipaddress.IPv4Address('1.2.3.4'), 3),
                ],
                zone: error,
            },
            put_subdomain_ipv4_result={('', 'example.com'): None,
                                       ('foo', 'example.com'): None,
                                       ('foo.bar', 'example.net'): None},
        )
        updater.init_hosts_and_zones(
            "example.com foo.bar.example.net foo.example.com"
        )
        with pytest.raises(error):
            updater.publish_ipv4(ipaddress.IPv4Address('5.6.7.8'))

        assert updater.fetch_zone_ipv4s_calls == [
            'example.com',
            'example.net',
        ]
        assert updater.fetch_subdomain_ipv4s_calls == []
        calls = [
            ('example.com', {
                '': ([ipaddress.IPv4Address('5.6.7.8')], 1),
                'foo': ([ipaddress.IPv4Address('5.6.7.8')], 2),
            }),
            ('example.net', {
                'foo.bar': ([ipaddress.IPv4Address('5.6.7.8')], 3),
            }),
        ]
        assert updater.put_zone_ipv4s_calls == [
            c for c in calls if c[0] != zone
        ]
        calls = [
            ('', 'example.com', ipaddress.IPv4Address('5.6.7.8'), 1),
            ('foo', 'example.com', ipaddress.IPv4Address('5.6.7.8'), 2),
            ('foo.bar', 'example.net', ipaddress.IPv4Address('5.6.7.8'), 3),
        ]
        assert updater.put_subdomain_ipv4_calls == [
            c for c in calls if c[1] != zone
        ]

    def test_put_zone_and_subdomain_not_implemented(self, empty_addrfile,
                                                    data_dir):
        """Test that FatalPublishError is raised when neither put_zone_ipv4s
        nor put_subdomain_ipv4 are implemented"""
        updater = doubles.MockTwoWayZoneUpdater(
            'test_updater', empty_addrfile, data_dir,
            fetch_zone_ipv4s_result={
                'example.com': [
                    ('', ipaddress.IPv4Address('1.2.3.4'), 1),
                    ('foo', ipaddress.IPv4Address('1.2.3.4'), 2),
                ],
                'example.net': [
                    ('foo.bar', ipaddress.IPv4Address('1.2.3.4'), 3),
                ],
            },
        )
        updater.init_hosts_and_zones(
            "example.com foo.bar.example.net foo.example.com"
        )
        with pytest.raises(FatalPublishError):
            updater.publish_ipv4(ipaddress.IPv4Address('5.6.7.8'))


class TestFetchZoneIPv4sNotImplemented:
    def test_put_subdomain_preferred(self, empty_addrfile, data_dir):
        """Test that put_subdomain_ipv4 is used when fetch_zone_ipv4s is not
        implemented"""
        updater = doubles.MockTwoWayZoneUpdater(
            'test_updater', empty_addrfile, data_dir,
            fetch_subdomain_ipv4s_result={
                ('', 'example.com'): [
                    (ipaddress.IPv4Address('1.2.3.4'), 1),
                ],
                ('foo', 'example.com'): [
                    (ipaddress.IPv4Address('1.2.3.4'), 2),
                ],
                ('foo.bar', 'example.net'): [
                    (ipaddress.IPv4Address('1.2.3.4'), 3),
                ],
            },
            put_zone_ipv4s_result={'example.com': None,
                                   'example.net': None},
            put_subdomain_ipv4_result={('', 'example.com'): None,
                                       ('foo', 'example.com'): None,
                                       ('foo.bar', 'example.net'): None},
        )
        updater.init_hosts_and_zones(
            "example.com foo.bar.example.net foo.example.com"
        )
        updater.publish_ipv4(ipaddress.IPv4Address('5.6.7.8'))

        assert updater.fetch_zone_ipv4s_calls == [
            'example.com',
            'example.net',
        ]
        assert updater.fetch_subdomain_ipv4s_calls == [
            ('', 'example.com'),
            ('foo', 'example.com'),
            ('foo.bar', 'example.net'),
        ]
        assert updater.put_zone_ipv4s_calls == []
        assert updater.put_subdomain_ipv4_calls == [
            ('', 'example.com', ipaddress.IPv4Address('5.6.7.8'), 1),
            ('foo', 'example.com', ipaddress.IPv4Address('5.6.7.8'), 2),
            ('foo.bar', 'example.net', ipaddress.IPv4Address('5.6.7.8'), 3),
        ]

    def test_put_subdomain_not_implemented(self, empty_addrfile, data_dir):
        """Test that when fetch_zone_ipv4s is not implemented,
        put_subdomain_ipv4 not implemented is a FatalPublishError even if
        put_zone_ipv4s is implemented"""
        updater = doubles.MockTwoWayZoneUpdater(
            'test_updater', empty_addrfile, data_dir,
            fetch_subdomain_ipv4s_result={
                ('', 'example.com'): [
                    (ipaddress.IPv4Address('1.2.3.4'), 1),
                ],
                ('foo', 'example.com'): [
                    (ipaddress.IPv4Address('1.2.3.4'), 2),
                ],
                ('foo.bar', 'example.net'): [
                    (ipaddress.IPv4Address('1.2.3.4'), 3),
                ],
            },
            put_zone_ipv4s_result={'example.com': None,
                                   'example.net': None},
        )
        updater.init_hosts_and_zones(
            "example.com foo.bar.example.net foo.example.com"
        )
        with pytest.raises(FatalPublishError):
            updater.publish_ipv4(ipaddress.IPv4Address('5.6.7.8'))

    def test_multiple_records(self, empty_addrfile, data_dir):
        """Test that multiple records for a host from fetch_subdomain_ipv4s
        still lead to a single call to put_subdomain_ipv4"""
        updater = doubles.MockTwoWayZoneUpdater(
            'test_updater', empty_addrfile, data_dir,
            fetch_subdomain_ipv4s_result={
                # Different TTL within an RR set is not valid, but some
                # providers may support it anyway. We take the minimum.
                ('', 'example.com'): [
                    (ipaddress.IPv4Address('1.2.3.4'), 1),
                    (ipaddress.IPv4Address('3.4.5.6'), 11),
                ],
                ('foo', 'example.com'): [
                    (ipaddress.IPv4Address('1.2.3.4'), 2),
                    (ipaddress.IPv4Address('3.4.5.6'), 22),
                ],
                ('foo.bar', 'example.net'): [
                    (ipaddress.IPv4Address('1.2.3.4'), 3),
                    (ipaddress.IPv4Address('3.4.5.6'), 33),
                ],
            },
            put_subdomain_ipv4_result={('', 'example.com'): None,
                                       ('foo', 'example.com'): None,
                                       ('foo.bar', 'example.net'): None},
        )
        updater.init_hosts_and_zones(
            "example.com foo.bar.example.net foo.example.com"
        )
        updater.publish_ipv4(ipaddress.IPv4Address('5.6.7.8'))

        assert updater.fetch_zone_ipv4s_calls == [
            'example.com',
            'example.net',
        ]
        assert updater.fetch_subdomain_ipv4s_calls == [
            ('', 'example.com'),
            ('foo', 'example.com'),
            ('foo.bar', 'example.net'),
        ]
        assert updater.put_zone_ipv4s_calls == []
        assert updater.put_subdomain_ipv4_calls == [
            ('', 'example.com', ipaddress.IPv4Address('5.6.7.8'), 1),
            ('foo', 'example.com', ipaddress.IPv4Address('5.6.7.8'), 2),
            ('foo.bar', 'example.net', ipaddress.IPv4Address('5.6.7.8'), 3),
        ]

    @pytest.mark.parametrize(('subdomain', 'zone', 'error'), [
        ('', 'example.com', PublishError),
        ('', 'example.com', FatalPublishError),
        ('foo', 'example.com', PublishError),
        ('foo', 'example.com', FatalPublishError),
        ('foo.bar', 'example.net', PublishError),
        ('foo.bar', 'example.net', FatalPublishError),
    ])
    def test_fetch_subdomain_publish_error(self, empty_addrfile, data_dir,
                                           subdomain, zone, error):
        """Test that remaining hosts are still updated when
        fetch_subdomain_ipv4s raises a PublishError for one of them"""
        updater = doubles.MockTwoWayZoneUpdater(
            'test_updater', empty_addrfile, data_dir,
            fetch_subdomain_ipv4s_result={
                ('', 'example.com'): [
                    (ipaddress.IPv4Address('1.2.3.4'), 1),
                ],
                ('foo', 'example.com'): [
                    (ipaddress.IPv4Address('1.2.3.4'), 2),
                ],
                ('foo.bar', 'example.net'): [
                    (ipaddress.IPv4Address('1.2.3.4'), 3),
                ],
                (subdomain, zone): error,
            },
            put_subdomain_ipv4_result={('', 'example.com'): None,
                                       ('foo', 'example.com'): None,
                                       ('foo.bar', 'example.net'): None},
        )
        updater.init_hosts_and_zones(
            "example.com foo.bar.example.net foo.example.com"
        )
        with pytest.raises(error):
            updater.publish_ipv4(ipaddress.IPv4Address('5.6.7.8'))

        assert updater.fetch_zone_ipv4s_calls == [
            'example.com',
            'example.net',
        ]
        assert updater.fetch_subdomain_ipv4s_calls == [
            ('', 'example.com'),
            ('foo', 'example.com'),
            ('foo.bar', 'example.net'),
        ]
        assert updater.put_zone_ipv4s_calls == []
        calls = [
            ('', 'example.com', ipaddress.IPv4Address('5.6.7.8'), 1),
            ('foo', 'example.com', ipaddress.IPv4Address('5.6.7.8'), 2),
            ('foo.bar', 'example.net', ipaddress.IPv4Address('5.6.7.8'), 3),
        ]
        assert updater.put_subdomain_ipv4_calls == [
            c for c in calls if c[0:2] != (subdomain, zone)
        ]

    @pytest.mark.parametrize(('subdomain', 'zone', 'error'), [
        ('', 'example.com', PublishError),
        ('', 'example.com', FatalPublishError),
        ('foo', 'example.com', PublishError),
        ('foo', 'example.com', FatalPublishError),
        ('foo.bar', 'example.net', PublishError),
        ('foo.bar', 'example.net', FatalPublishError),
    ])
    def test_put_subdomain_publish_error(self, empty_addrfile, data_dir,
                                         subdomain, zone, error):
        """Test that remaining hosts are still updated when put_subdomain_ipv4
        raises a PublishError for one of them"""
        updater = doubles.MockTwoWayZoneUpdater(
            'test_updater', empty_addrfile, data_dir,
            fetch_subdomain_ipv4s_result={
                ('', 'example.com'): [
                    (ipaddress.IPv4Address('1.2.3.4'), 1),
                ],
                ('foo', 'example.com'): [
                    (ipaddress.IPv4Address('1.2.3.4'), 2),
                ],
                ('foo.bar', 'example.net'): [
                    (ipaddress.IPv4Address('1.2.3.4'), 3),
                ],
            },
            put_subdomain_ipv4_result={('', 'example.com'): None,
                                       ('foo', 'example.com'): None,
                                       ('foo.bar', 'example.net'): None,
                                       (subdomain, zone): error},
        )
        updater.init_hosts_and_zones(
            "example.com foo.bar.example.net foo.example.com"
        )
        with pytest.raises(error):
            updater.publish_ipv4(ipaddress.IPv4Address('5.6.7.8'))

        assert updater.fetch_zone_ipv4s_calls == [
            'example.com',
            'example.net',
        ]
        assert updater.fetch_subdomain_ipv4s_calls == [
            ('', 'example.com'),
            ('foo', 'example.com'),
            ('foo.bar', 'example.net'),
        ]
        assert updater.put_zone_ipv4s_calls == []
        assert updater.put_subdomain_ipv4_calls == [
            ('', 'example.com', ipaddress.IPv4Address('5.6.7.8'), 1),
            ('foo', 'example.com', ipaddress.IPv4Address('5.6.7.8'), 2),
            ('foo.bar', 'example.net', ipaddress.IPv4Address('5.6.7.8'), 3),
        ]


def test_fetch_zone_and_subdomain_ipv4_not_implemented(empty_addrfile,
                                                       data_dir):
    """Test that it's a FatalPublishError when neither fetch_zone_ipv4s nor
    fetch_subdomain_ipv4s are implemented"""
    updater = doubles.MockTwoWayZoneUpdater(
        'test_updater', empty_addrfile, data_dir,
        put_zone_ipv4s_result={'example.com': None,
                               'example.net': None},
        put_subdomain_ipv4_result={('', 'example.com'): None,
                                   ('foo', 'example.com'): None,
                                   ('foo.bar', 'example.net'): None},
    )
    updater.init_hosts_and_zones(
        "example.com foo.bar.example.net foo.example.com"
    )
    with pytest.raises(FatalPublishError):
        updater.publish_ipv4(ipaddress.IPv4Address('5.6.7.8'))


class TestFetchZoneIPv6sImplemented:
    def test_fetch_and_put_zone_preferred(self, empty_addrfile, data_dir):
        """Test that fetch_zone_ipv6s and put_zone_ipv6s are preferred over
        fetch_subdomain_ipv6s and put_subdomain_ipv6 when all are
        implemented"""
        updater = doubles.MockTwoWayZoneUpdater(
            'test_updater', empty_addrfile, data_dir,
            fetch_zone_ipv6s_result={
                'example.com': [
                    ('', ipaddress.IPv6Address('1234::1'), 1),
                    ('foo', ipaddress.IPv6Address('1234::2'), 2),
                ],
                'example.net': [
                    ('foo.bar', ipaddress.IPv6Address('1234::3'), 3),
                ],
            },
            fetch_subdomain_ipv6s_result={
                ('', 'example.com'): [
                    (ipaddress.IPv6Address('1234::1'), 1),
                ],
                ('foo', 'example.com'): [
                    (ipaddress.IPv6Address('1234::2'), 2),
                ],
                ('foo.bar', 'example.net'): [
                    (ipaddress.IPv6Address('1234::3'), 3),
                ],
            },
            put_zone_ipv6s_result={'example.com': None,
                                   'example.net': None},
            put_subdomain_ipv6s_result={('', 'example.com'): None,
                                        ('foo', 'example.com'): None,
                                        ('foo.bar', 'example.net'): None},
        )
        updater.init_hosts_and_zones(
            "example.com foo.bar.example.net foo.example.com"
        )
        updater.publish_ipv6(ipaddress.IPv6Network('5678::/64'))

        assert updater.fetch_zone_ipv6s_calls == [
            'example.com',
            'example.net',
        ]
        assert updater.fetch_subdomain_ipv6s_calls == []
        assert updater.put_zone_ipv6s_calls == [
            ('example.com', {
                '': ([ipaddress.IPv6Address('5678::1')], 1),
                'foo': ([ipaddress.IPv6Address('5678::2')], 2),
            }),
            ('example.net', {
                'foo.bar': ([ipaddress.IPv6Address('5678::3')], 3),
            }),
        ]
        assert updater.put_subdomain_ipv6s_calls == []

    def test_extra_records_untouched(self, empty_addrfile, data_dir):
        """Test extra records from fetch_zone_ipv6s are passed to
        put_zone_ipv6s untouched"""
        updater = doubles.MockTwoWayZoneUpdater(
            'test_updater', empty_addrfile, data_dir,
            fetch_zone_ipv6s_result={
                'example.com': [
                    ('', ipaddress.IPv6Address('1234::1'), 1),
                    ('foo', ipaddress.IPv6Address('1234::2'), 2),
                    ('bar', ipaddress.IPv6Address('1234::4'), 4),
                ],
                'example.net': [
                    ('foo.bar', ipaddress.IPv6Address('1234::3'), 3),
                    ('baz', ipaddress.IPv6Address('3456::5'), 5),
                ],
            },
            put_zone_ipv6s_result={'example.com': None,
                                   'example.net': None},
        )
        updater.init_hosts_and_zones(
            "example.com foo.bar.example.net foo.example.com"
        )
        updater.publish_ipv6(ipaddress.IPv6Network('5678::/64'))

        assert updater.fetch_zone_ipv6s_calls == [
            'example.com',
            'example.net',
        ]
        assert updater.fetch_subdomain_ipv6s_calls == []
        assert updater.put_zone_ipv6s_calls == [
            ('example.com', {
                '': ([ipaddress.IPv6Address('5678::1')], 1),
                'foo': ([ipaddress.IPv6Address('5678::2')], 2),
                'bar': ([ipaddress.IPv6Address('1234::4')], 4),
            }),
            ('example.net', {
                'foo.bar': ([ipaddress.IPv6Address('5678::3')], 3),
                'baz': ([ipaddress.IPv6Address('3456::5')], 5),
            }),
        ]
        assert updater.put_subdomain_ipv6s_calls == []

    def test_missing_record(self, empty_addrfile, data_dir):
        """Test missing record from fetch_zone_ipv6s is PublishError but other
        records still updated"""
        updater = doubles.MockTwoWayZoneUpdater(
            'test_updater', empty_addrfile, data_dir,
            fetch_zone_ipv6s_result={
                'example.com': [
                    ('foo', ipaddress.IPv6Address('1234::2'), 2),
                ],
                'example.net': [
                    ('foo.bar', ipaddress.IPv6Address('1234::3'), 3),
                ],
            },
            put_zone_ipv6s_result={'example.com': None,
                                   'example.net': None},
        )
        updater.init_hosts_and_zones(
            "example.com foo.bar.example.net foo.example.com"
        )
        with pytest.raises(PublishError):
            updater.publish_ipv6(ipaddress.IPv6Network('5678::/64'))

        assert updater.fetch_zone_ipv6s_calls == [
            'example.com',
            'example.net',
        ]
        assert updater.fetch_subdomain_ipv6s_calls == []
        assert updater.put_zone_ipv6s_calls == [
            ('example.com', {
                'foo': ([ipaddress.IPv6Address('5678::2')], 2),
            }),
            ('example.net', {
                'foo.bar': ([ipaddress.IPv6Address('5678::3')], 3),
            }),
        ]
        assert updater.put_subdomain_ipv6s_calls == []

    def test_multiple_records(self, empty_addrfile, data_dir):
        """Test when multiple records are set for a host, they are all
        replaced, but multiple records on irrelevant hosts are left alone."""
        updater = doubles.MockTwoWayZoneUpdater(
            'test_updater', empty_addrfile, data_dir,
            fetch_zone_ipv6s_result={
                'example.com': [
                    # Different TTL within an RR set is not valid, but some
                    # providers may support it anyway. We take the minimum.
                    ('', ipaddress.IPv6Address('1234::1'), 1),
                    ('', ipaddress.IPv6Address('2345::1'), 11),
                    ('foo', ipaddress.IPv6Address('1234::2'), 2),
                    ('foo', ipaddress.IPv6Address('1234::22'), 22),
                    ('bar', ipaddress.IPv6Address('1234::4'), 4),
                    ('bar', ipaddress.IPv6Address('2345::4'), 44),
                ],
                'example.net': [
                    ('foo.bar', ipaddress.IPv6Address('1234::3'), 3),
                    ('foo.bar', ipaddress.IPv6Address('2345::33'), 33),
                    ('baz', ipaddress.IPv6Address('3456::5'), 5),
                    ('baz', ipaddress.IPv6Address('3456::55'), 55),
                ],
            },
            put_zone_ipv6s_result={'example.com': None,
                                   'example.net': None},
        )
        updater.init_hosts_and_zones(
            "example.com foo.bar.example.net foo.example.com"
        )
        updater.publish_ipv6(ipaddress.IPv6Network('5678::/63'))

        assert updater.fetch_zone_ipv6s_calls == [
            'example.com',
            'example.net',
        ]
        assert updater.fetch_subdomain_ipv6s_calls == []
        assert updater.put_zone_ipv6s_calls == [
            ('example.com', {
                '': ([ipaddress.IPv6Address('5678::1')], 1),
                'foo': ([ipaddress.IPv6Address('5678::2'),
                         ipaddress.IPv6Address('5678::22')], 2),
                'bar': ([ipaddress.IPv6Address('1234::4'),
                         ipaddress.IPv6Address('2345::4')], 4),
            }),
            ('example.net', {
                'foo.bar': ([ipaddress.IPv6Address('5678::3'),
                             ipaddress.IPv6Address('5678::33')], 3),
                'baz': ([ipaddress.IPv6Address('3456::5'),
                         ipaddress.IPv6Address('3456::55')], 5),
            }),
        ]
        assert updater.put_subdomain_ipv6s_calls == []

    @pytest.mark.parametrize(('zone', 'error'), [
        ('example.com', PublishError),
        ('example.com', FatalPublishError),
        ('example.net', PublishError),
        ('example.net', FatalPublishError),
    ])
    def test_put_zone_publish_error(self, empty_addrfile, data_dir,
                                    zone, error):
        """Test other zones are still published after PublishError for one
        zone's put_zone_ipv6s"""
        updater = doubles.MockTwoWayZoneUpdater(
            'test_updater', empty_addrfile, data_dir,
            fetch_zone_ipv6s_result={
                'example.com': [
                    ('', ipaddress.IPv6Address('1234::1'), 1),
                    ('foo', ipaddress.IPv6Address('1234::2'), 2),
                ],
                'example.net': [
                    ('foo.bar', ipaddress.IPv6Address('1234::3'), 3),
                ],
            },
            put_zone_ipv6s_result={'example.com': None,
                                   'example.net': None,
                                   zone: error},
        )
        updater.init_hosts_and_zones(
            "example.com foo.bar.example.net foo.example.com"
        )
        with pytest.raises(error):
            updater.publish_ipv6(ipaddress.IPv6Network('5678::/64'))

        assert updater.fetch_zone_ipv6s_calls == [
            'example.com',
            'example.net',
        ]
        assert updater.fetch_subdomain_ipv6s_calls == []
        assert updater.put_zone_ipv6s_calls == [
            ('example.com', {
                '': ([ipaddress.IPv6Address('5678::1')], 1),
                'foo': ([ipaddress.IPv6Address('5678::2')], 2),
            }),
            ('example.net', {
                'foo.bar': ([ipaddress.IPv6Address('5678::3')], 3),
            }),
        ]
        assert updater.put_subdomain_ipv6s_calls == []

    @pytest.mark.parametrize(('zone', 'error'), [
        ('example.com', PublishError),
        ('example.com', FatalPublishError),
        ('example.net', PublishError),
        ('example.net', FatalPublishError),
    ])
    def test_fetch_zone_publish_error(self, empty_addrfile, data_dir,
                                      zone, error):
        """Test other zones still updated when fetch_zone_ipv6s raises
        PublishError for one zone"""
        updater = doubles.MockTwoWayZoneUpdater(
            'test_updater', empty_addrfile, data_dir,
            fetch_zone_ipv6s_result={
                'example.com': [
                    ('', ipaddress.IPv6Address('1234::1'), 1),
                    ('foo', ipaddress.IPv6Address('1234::2'), 2),
                ],
                'example.net': [
                    ('foo.bar', ipaddress.IPv6Address('1234::3'), 3),
                ],
                zone: error,
            },
            put_zone_ipv6s_result={'example.com': None,
                                   'example.net': None},
        )
        updater.init_hosts_and_zones(
            "example.com foo.bar.example.net foo.example.com"
        )
        with pytest.raises(error):
            updater.publish_ipv6(ipaddress.IPv6Network('5678::/64'))

        assert updater.fetch_zone_ipv6s_calls == [
            'example.com',
            'example.net',
        ]
        assert updater.fetch_subdomain_ipv6s_calls == []
        calls = [
            ('example.com', {
                '': ([ipaddress.IPv6Address('5678::1')], 1),
                'foo': ([ipaddress.IPv6Address('5678::2')], 2),
            }),
            ('example.net', {
                'foo.bar': ([ipaddress.IPv6Address('5678::3')], 3),
            }),
        ]
        assert updater.put_zone_ipv6s_calls == [
            c for c in calls if c[0] != zone
        ]
        assert updater.put_subdomain_ipv6s_calls == []

    def test_put_zone_not_implemented(self, empty_addrfile, data_dir):
        """Test that put_subdomain_ipv6s is used when fetch_zone_ipv6s is
        implemented but put_zone_ipv6 is not"""
        updater = doubles.MockTwoWayZoneUpdater(
            'test_updater', empty_addrfile, data_dir,
            fetch_zone_ipv6s_result={
                'example.com': [
                    ('', ipaddress.IPv6Address('1234::1'), 1),
                    ('foo', ipaddress.IPv6Address('1234::2'), 2),
                ],
                'example.net': [
                    ('foo.bar', ipaddress.IPv6Address('1234::3'), 3),
                ],
            },
            put_subdomain_ipv6s_result={('', 'example.com'): None,
                                        ('foo', 'example.com'): None,
                                        ('foo.bar', 'example.net'): None},
        )
        updater.init_hosts_and_zones(
            "example.com foo.bar.example.net foo.example.com"
        )
        updater.publish_ipv6(ipaddress.IPv6Network('5678::/64'))

        assert updater.fetch_zone_ipv6s_calls == [
            'example.com',
            'example.net',
        ]
        assert updater.fetch_subdomain_ipv6s_calls == []
        assert updater.put_zone_ipv6s_calls == [
            ('example.com', {
                '': ([ipaddress.IPv6Address('5678::1')], 1),
                'foo': ([ipaddress.IPv6Address('5678::2')], 2),
            }),
            ('example.net', {
                'foo.bar': ([ipaddress.IPv6Address('5678::3')], 3),
            }),
        ]
        assert updater.put_subdomain_ipv6s_calls == [
            ('', 'example.com', [ipaddress.IPv6Address('5678::1')], 1),
            ('foo', 'example.com', [ipaddress.IPv6Address('5678::2')], 2),
            ('foo.bar', 'example.net', [ipaddress.IPv6Address('5678::3')], 3),
        ]

    def test_put_zone_not_implemented_extra_records(self, empty_addrfile,
                                                    data_dir):
        """Test that put_subdomain_ipv6s is not called for extra records
        returned by fetch_zone_ipv6s"""
        updater = doubles.MockTwoWayZoneUpdater(
            'test_updater', empty_addrfile, data_dir,
            fetch_zone_ipv6s_result={
                'example.com': [
                    ('', ipaddress.IPv6Address('1234::1'), 1),
                    ('foo', ipaddress.IPv6Address('1234::2'), 2),
                    ('bar', ipaddress.IPv6Address('1234::4'), 4),
                ],
                'example.net': [
                    ('foo.bar', ipaddress.IPv6Address('1234::3'), 3),
                    ('baz', ipaddress.IPv6Address('3456::5'), 5),
                ],
            },
            put_subdomain_ipv6s_result={('', 'example.com'): None,
                                        ('foo', 'example.com'): None,
                                        ('foo.bar', 'example.net'): None},
        )
        updater.init_hosts_and_zones(
            "example.com foo.bar.example.net foo.example.com"
        )
        updater.publish_ipv6(ipaddress.IPv6Network('5678::/64'))

        assert updater.fetch_zone_ipv6s_calls == [
            'example.com',
            'example.net',
        ]
        assert updater.fetch_subdomain_ipv6s_calls == []
        assert updater.put_zone_ipv6s_calls == [
            ('example.com', {
                '': ([ipaddress.IPv6Address('5678::1')], 1),
                'foo': ([ipaddress.IPv6Address('5678::2')], 2),
                'bar': ([ipaddress.IPv6Address('1234::4')], 4),
            }),
            ('example.net', {
                'foo.bar': ([ipaddress.IPv6Address('5678::3')], 3),
                'baz': ([ipaddress.IPv6Address('3456::5')], 5),
            }),
        ]
        assert updater.put_subdomain_ipv6s_calls == [
            ('', 'example.com', [ipaddress.IPv6Address('5678::1')], 1),
            ('foo', 'example.com', [ipaddress.IPv6Address('5678::2')], 2),
            ('foo.bar', 'example.net', [ipaddress.IPv6Address('5678::3')], 3),
        ]

    def test_put_zone_not_implemented_missing_record(self, empty_addrfile,
                                                     data_dir):
        """Test missing record from fetch_zone_ipv6s is PublishError but other
        records still updated with put_subdomain_ipv6s when put_zone_ipv6s not
        implemented"""
        updater = doubles.MockTwoWayZoneUpdater(
            'test_updater', empty_addrfile, data_dir,
            fetch_zone_ipv6s_result={
                'example.com': [
                    ('foo', ipaddress.IPv6Address('1234::2'), 2),
                ],
                'example.net': [
                    ('foo.bar', ipaddress.IPv6Address('1234::3'), 3),
                ],
            },
            put_subdomain_ipv6s_result={('foo', 'example.com'): None,
                                        ('foo.bar', 'example.net'): None},
        )
        updater.init_hosts_and_zones(
            "example.com foo.bar.example.net foo.example.com"
        )
        with pytest.raises(PublishError):
            updater.publish_ipv6(ipaddress.IPv6Network('5678::/64'))

        assert updater.fetch_zone_ipv6s_calls == [
            'example.com',
            'example.net',
        ]
        assert updater.fetch_subdomain_ipv6s_calls == []
        assert updater.put_zone_ipv6s_calls == [
            ('example.com', {
                'foo': ([ipaddress.IPv6Address('5678::2')], 2),
            }),
            ('example.net', {
                'foo.bar': ([ipaddress.IPv6Address('5678::3')], 3),
            }),
        ]
        assert updater.put_subdomain_ipv6s_calls == [
            ('foo', 'example.com', [ipaddress.IPv6Address('5678::2')], 2),
            ('foo.bar', 'example.net', [ipaddress.IPv6Address('5678::3')], 3),
        ]

    def test_put_zone_not_implemented_multiple_records(
        self, empty_addrfile, data_dir
    ):
        """Test when multiple records are set for a host and put_zone_ipv6s is
        not implemented, put_subdomain_ipv6s is called exactly once for each
        relevant host and not at all for irrelevant hosts"""
        updater = doubles.MockTwoWayZoneUpdater(
            'test_updater', empty_addrfile, data_dir,
            fetch_zone_ipv6s_result={
                'example.com': [
                    # Different TTL within an RR set is not valid, but some
                    # providers may support it anyway. We take the minimum.
                    ('', ipaddress.IPv6Address('1234::1'), 1),
                    ('', ipaddress.IPv6Address('2345::1'), 11),
                    ('foo', ipaddress.IPv6Address('1234::2'), 2),
                    ('foo', ipaddress.IPv6Address('1234::22'), 22),
                    ('bar', ipaddress.IPv6Address('1234::4'), 4),
                    ('bar', ipaddress.IPv6Address('2345::4'), 44),
                ],
                'example.net': [
                    ('foo.bar', ipaddress.IPv6Address('1234::3'), 3),
                    ('foo.bar', ipaddress.IPv6Address('2345::33'), 33),
                    ('baz', ipaddress.IPv6Address('3456::5'), 5),
                    ('baz', ipaddress.IPv6Address('3456::55'), 55),
                ],
            },
            put_subdomain_ipv6s_result={('', 'example.com'): None,
                                        ('foo', 'example.com'): None,
                                        ('foo.bar', 'example.net'): None},
        )
        updater.init_hosts_and_zones(
            "example.com foo.bar.example.net foo.example.com"
        )
        updater.publish_ipv6(ipaddress.IPv6Network('5678::/64'))

        assert updater.fetch_zone_ipv6s_calls == [
            'example.com',
            'example.net',
        ]
        assert updater.fetch_subdomain_ipv6s_calls == []
        assert updater.put_zone_ipv6s_calls == [
            ('example.com', {
                '': ([ipaddress.IPv6Address('5678::1')], 1),
                'foo': ([ipaddress.IPv6Address('5678::2'),
                         ipaddress.IPv6Address('5678::22')], 2),
                'bar': ([ipaddress.IPv6Address('1234::4'),
                         ipaddress.IPv6Address('2345::4')], 4),
            }),
            ('example.net', {
                'foo.bar': ([ipaddress.IPv6Address('5678::3'),
                             ipaddress.IPv6Address('5678::33')], 3),
                'baz': ([ipaddress.IPv6Address('3456::5'),
                         ipaddress.IPv6Address('3456::55')], 5),
            }),
        ]
        assert updater.put_subdomain_ipv6s_calls == [
            ('', 'example.com', [ipaddress.IPv6Address('5678::1')], 1),
            ('foo', 'example.com', [ipaddress.IPv6Address('5678::2'),
                                    ipaddress.IPv6Address('5678::22')], 2),
            ('foo.bar', 'example.net', [ipaddress.IPv6Address('5678::3'),
                                        ipaddress.IPv6Address('5678::33')], 3),
        ]

    @pytest.mark.parametrize(('subdomain', 'zone', 'error'), [
        ('', 'example.com', PublishError),
        ('', 'example.com', FatalPublishError),
        ('foo', 'example.com', PublishError),
        ('foo', 'example.com', FatalPublishError),
        ('foo.bar', 'example.net', PublishError),
        ('foo.bar', 'example.net', FatalPublishError),
    ])
    def test_put_zone_not_implemented_put_subdomain_publish_error(
        self, empty_addrfile, data_dir, subdomain, zone, error
    ):
        """Test that remaining subdomains and zones are still published when
        put_subdomain_ipv6s has a PublishError"""
        updater = doubles.MockTwoWayZoneUpdater(
            'test_updater', empty_addrfile, data_dir,
            fetch_zone_ipv6s_result={
                'example.com': [
                    ('', ipaddress.IPv6Address('1234::1'), 1),
                    ('foo', ipaddress.IPv6Address('1234::2'), 2),
                ],
                'example.net': [
                    ('foo.bar', ipaddress.IPv6Address('1234::3'), 3),
                ],
            },
            put_subdomain_ipv6s_result={('', 'example.com'): None,
                                        ('foo', 'example.com'): None,
                                        ('foo.bar', 'example.net'): None,
                                        (subdomain, zone): error},
        )
        updater.init_hosts_and_zones(
            "example.com foo.bar.example.net foo.example.com"
        )
        with pytest.raises(error):
            updater.publish_ipv6(ipaddress.IPv6Network('5678::/64'))

        assert updater.fetch_zone_ipv6s_calls == [
            'example.com',
            'example.net',
        ]
        assert updater.fetch_subdomain_ipv6s_calls == []
        assert updater.put_zone_ipv6s_calls == [
            ('example.com', {
                '': ([ipaddress.IPv6Address('5678::1')], 1),
                'foo': ([ipaddress.IPv6Address('5678::2')], 2),
            }),
            ('example.net', {
                'foo.bar': ([ipaddress.IPv6Address('5678::3')], 3),
            }),
        ]
        assert updater.put_subdomain_ipv6s_calls == [
            ('', 'example.com', [ipaddress.IPv6Address('5678::1')], 1),
            ('foo', 'example.com', [ipaddress.IPv6Address('5678::2')], 2),
            ('foo.bar', 'example.net', [ipaddress.IPv6Address('5678::3')], 3),
        ]

    @pytest.mark.parametrize(('zone', 'error'), [
        ('example.com', PublishError),
        ('example.com', FatalPublishError),
        ('example.net', PublishError),
        ('example.net', FatalPublishError),
    ])
    def test_put_zone_not_implemented_fetch_zone_publish_error(
        self, empty_addrfile, data_dir, zone, error
    ):
        """Test other zones still updated when fetch_zone_ipv4s raises
        PublishError for one zone and put_zone_ipv4s not implemented"""
        updater = doubles.MockTwoWayZoneUpdater(
            'test_updater', empty_addrfile, data_dir,
            fetch_zone_ipv6s_result={
                'example.com': [
                    ('', ipaddress.IPv6Address('1234::1'), 1),
                    ('foo', ipaddress.IPv6Address('1234::2'), 2),
                ],
                'example.net': [
                    ('foo.bar', ipaddress.IPv6Address('1234::3'), 3),
                ],
                zone: error,
            },
            put_subdomain_ipv6s_result={('', 'example.com'): None,
                                        ('foo', 'example.com'): None,
                                        ('foo.bar', 'example.net'): None},
        )
        updater.init_hosts_and_zones(
            "example.com foo.bar.example.net foo.example.com"
        )
        with pytest.raises(error):
            updater.publish_ipv6(ipaddress.IPv6Network('5678::/64'))

        assert updater.fetch_zone_ipv6s_calls == [
            'example.com',
            'example.net',
        ]
        assert updater.fetch_subdomain_ipv6s_calls == []
        calls = [
            ('example.com', {
                '': ([ipaddress.IPv6Address('5678::1')], 1),
                'foo': ([ipaddress.IPv6Address('5678::2')], 2),
            }),
            ('example.net', {
                'foo.bar': ([ipaddress.IPv6Address('5678::3')], 3),
            }),
        ]
        assert updater.put_zone_ipv6s_calls == [
            c for c in calls if c[0] != zone
        ]
        calls = [
            ('', 'example.com', [ipaddress.IPv6Address('5678::1')], 1),
            ('foo', 'example.com', [ipaddress.IPv6Address('5678::2')], 2),
            ('foo.bar', 'example.net', [ipaddress.IPv6Address('5678::3')], 3),
        ]
        assert updater.put_subdomain_ipv6s_calls == [
            c for c in calls if c[1] != zone
        ]

    def test_put_zone_and_subdomain_not_implemented(self, empty_addrfile,
                                                    data_dir):
        """Test that FatalPublishError is raised when neither put_zone_ipv6s
        nor put_subdomain_ipv6s are implemented"""
        updater = doubles.MockTwoWayZoneUpdater(
            'test_updater', empty_addrfile, data_dir,
            fetch_zone_ipv6s_result={
                'example.com': [
                    ('', ipaddress.IPv6Address('1234::1'), 1),
                    ('foo', ipaddress.IPv6Address('1234::2'), 2),
                ],
                'example.net': [
                    ('foo.bar', ipaddress.IPv6Address('1234::3'), 3),
                ],
            },
        )
        updater.init_hosts_and_zones(
            "example.com foo.bar.example.net foo.example.com"
        )
        with pytest.raises(FatalPublishError):
            updater.publish_ipv6(ipaddress.IPv6Network('5678::/64'))


class TestFetchZoneIPv6sNotImplemented:
    def test_put_subdomain_preferred(self, empty_addrfile, data_dir):
        """Test that put_subdomain_ipv6s is used when fetch_zone_ipv6s is not
        implemented"""
        updater = doubles.MockTwoWayZoneUpdater(
            'test_updater', empty_addrfile, data_dir,
            fetch_subdomain_ipv6s_result={
                ('', 'example.com'): [
                    (ipaddress.IPv6Address('1234::1'), 1),
                ],
                ('foo', 'example.com'): [
                    (ipaddress.IPv6Address('1234::2'), 2),
                ],
                ('foo.bar', 'example.net'): [
                    (ipaddress.IPv6Address('1234::3'), 3),
                ],
            },
            put_zone_ipv6s_result={'example.com': None,
                                   'example.net': None},
            put_subdomain_ipv6s_result={('', 'example.com'): None,
                                        ('foo', 'example.com'): None,
                                        ('foo.bar', 'example.net'): None},
        )
        updater.init_hosts_and_zones(
            "example.com foo.bar.example.net foo.example.com"
        )
        updater.publish_ipv6(ipaddress.IPv6Network('5678::/64'))

        assert updater.fetch_zone_ipv6s_calls == [
            'example.com',
            'example.net',
        ]
        assert updater.fetch_subdomain_ipv6s_calls == [
            ('', 'example.com'),
            ('foo', 'example.com'),
            ('foo.bar', 'example.net'),
        ]
        assert updater.put_zone_ipv6s_calls == []
        assert updater.put_subdomain_ipv6s_calls == [
            ('', 'example.com', [ipaddress.IPv6Address('5678::1')], 1),
            ('foo', 'example.com', [ipaddress.IPv6Address('5678::2')], 2),
            ('foo.bar', 'example.net', [ipaddress.IPv6Address('5678::3')], 3),
        ]

    def test_put_subdomain_not_implemented(self, empty_addrfile, data_dir):
        """Test that when fetch_zone_ipv6s is not implemented,
        put_subdomain_ipv6s not implemented is a FatalPublishError even if
        put_zone_ipv6s is implemented"""
        updater = doubles.MockTwoWayZoneUpdater(
            'test_updater', empty_addrfile, data_dir,
            fetch_subdomain_ipv6s_result={
                ('', 'example.com'): [
                    (ipaddress.IPv6Address('1234::1'), 1),
                ],
                ('foo', 'example.com'): [
                    (ipaddress.IPv6Address('1234::2'), 2),
                ],
                ('foo.bar', 'example.net'): [
                    (ipaddress.IPv6Address('1234::3'), 3),
                ],
            },
            put_zone_ipv6s_result={'example.com': None,
                                   'example.net': None},
        )
        updater.init_hosts_and_zones(
            "example.com foo.bar.example.net foo.example.com"
        )
        with pytest.raises(FatalPublishError):
            updater.publish_ipv6(ipaddress.IPv6Network('5678::/64'))

    def test_multiple_records(self, empty_addrfile, data_dir):
        """Test that multiple records for a host from fetch_subdomain_ipv6s
        still lead to a single call to put_subdomain_ipv6s"""
        updater = doubles.MockTwoWayZoneUpdater(
            'test_updater', empty_addrfile, data_dir,
            fetch_subdomain_ipv6s_result={
                # Different TTL within an RR set is not valid, but some
                # providers may support it anyway. We take the minimum.
                ('', 'example.com'): [
                    (ipaddress.IPv6Address('1234::1'), 1),
                    (ipaddress.IPv6Address('3456::1'), 11),
                ],
                ('foo', 'example.com'): [
                    (ipaddress.IPv6Address('1234::2'), 2),
                    (ipaddress.IPv6Address('1234::22'), 22),
                ],
                ('foo.bar', 'example.net'): [
                    (ipaddress.IPv6Address('1234::3'), 3),
                    (ipaddress.IPv6Address('3456::33'), 33),
                ],
            },
            put_subdomain_ipv6s_result={('', 'example.com'): None,
                                        ('foo', 'example.com'): None,
                                        ('foo.bar', 'example.net'): None},
        )
        updater.init_hosts_and_zones(
            "example.com foo.bar.example.net foo.example.com"
        )
        updater.publish_ipv6(ipaddress.IPv6Network('5678::/64'))

        assert updater.fetch_zone_ipv6s_calls == [
            'example.com',
            'example.net',
        ]
        assert updater.fetch_subdomain_ipv6s_calls == [
            ('', 'example.com'),
            ('foo', 'example.com'),
            ('foo.bar', 'example.net'),
        ]
        assert updater.put_zone_ipv6s_calls == []
        assert updater.put_subdomain_ipv6s_calls == [
            ('', 'example.com', [ipaddress.IPv6Address('5678::1')], 1),
            ('foo', 'example.com', [ipaddress.IPv6Address('5678::2'),
                                    ipaddress.IPv6Address('5678::22')], 2),
            ('foo.bar', 'example.net', [ipaddress.IPv6Address('5678::3'),
                                        ipaddress.IPv6Address('5678::33')], 3),
        ]

    @pytest.mark.parametrize(('subdomain', 'zone', 'error'), [
        ('', 'example.com', PublishError),
        ('', 'example.com', FatalPublishError),
        ('foo', 'example.com', PublishError),
        ('foo', 'example.com', FatalPublishError),
        ('foo.bar', 'example.net', PublishError),
        ('foo.bar', 'example.net', FatalPublishError),
    ])
    def test_fetch_subdomain_publish_error(self, empty_addrfile, data_dir,
                                           subdomain, zone, error):
        """Test that remaining hosts are still updated when
        fetch_subdomain_ipv6s raises a PublishError for one of them"""
        updater = doubles.MockTwoWayZoneUpdater(
            'test_updater', empty_addrfile, data_dir,
            fetch_subdomain_ipv6s_result={
                ('', 'example.com'): [
                    (ipaddress.IPv6Address('1234::1'), 1),
                ],
                ('foo', 'example.com'): [
                    (ipaddress.IPv6Address('1234::2'), 2),
                ],
                ('foo.bar', 'example.net'): [
                    (ipaddress.IPv6Address('1234::3'), 3),
                ],
                (subdomain, zone): error,
            },
            put_subdomain_ipv6s_result={('', 'example.com'): None,
                                        ('foo', 'example.com'): None,
                                        ('foo.bar', 'example.net'): None},
        )
        updater.init_hosts_and_zones(
            "example.com foo.bar.example.net foo.example.com"
        )
        with pytest.raises(error):
            updater.publish_ipv6(ipaddress.IPv6Network('5678::/64'))

        assert updater.fetch_zone_ipv6s_calls == [
            'example.com',
            'example.net',
        ]
        assert updater.fetch_subdomain_ipv6s_calls == [
            ('', 'example.com'),
            ('foo', 'example.com'),
            ('foo.bar', 'example.net'),
        ]
        assert updater.put_zone_ipv6s_calls == []
        calls = [
            ('', 'example.com', [ipaddress.IPv6Address('5678::1')], 1),
            ('foo', 'example.com', [ipaddress.IPv6Address('5678::2')], 2),
            ('foo.bar', 'example.net', [ipaddress.IPv6Address('5678::3')], 3),
        ]
        assert updater.put_subdomain_ipv6s_calls == [
            c for c in calls if c[0:2] != (subdomain, zone)
        ]

    @pytest.mark.parametrize(('subdomain', 'zone', 'error'), [
        ('', 'example.com', PublishError),
        ('', 'example.com', FatalPublishError),
        ('foo', 'example.com', PublishError),
        ('foo', 'example.com', FatalPublishError),
        ('foo.bar', 'example.net', PublishError),
        ('foo.bar', 'example.net', FatalPublishError),
    ])
    def test_put_subdomain_publish_error(self, empty_addrfile, data_dir,
                                         subdomain, zone, error):
        """Test that remaining hosts are still updated when put_subdomain_ipv6s
        raises a PublishError for one of them"""
        updater = doubles.MockTwoWayZoneUpdater(
            'test_updater', empty_addrfile, data_dir,
            fetch_subdomain_ipv6s_result={
                ('', 'example.com'): [
                    (ipaddress.IPv6Address('1234::1'), 1),
                ],
                ('foo', 'example.com'): [
                    (ipaddress.IPv6Address('1234::2'), 2),
                ],
                ('foo.bar', 'example.net'): [
                    (ipaddress.IPv6Address('1234::3'), 3),
                ],
            },
            put_subdomain_ipv6s_result={('', 'example.com'): None,
                                        ('foo', 'example.com'): None,
                                        ('foo.bar', 'example.net'): None,
                                        (subdomain, zone): error},
        )
        updater.init_hosts_and_zones(
            "example.com foo.bar.example.net foo.example.com"
        )
        with pytest.raises(error):
            updater.publish_ipv6(ipaddress.IPv6Network('5678::/64'))

        assert updater.fetch_zone_ipv6s_calls == [
            'example.com',
            'example.net',
        ]
        assert updater.fetch_subdomain_ipv6s_calls == [
            ('', 'example.com'),
            ('foo', 'example.com'),
            ('foo.bar', 'example.net'),
        ]
        assert updater.put_zone_ipv6s_calls == []
        assert updater.put_subdomain_ipv6s_calls == [
            ('', 'example.com', [ipaddress.IPv6Address('5678::1')], 1),
            ('foo', 'example.com', [ipaddress.IPv6Address('5678::2')], 2),
            ('foo.bar', 'example.net', [ipaddress.IPv6Address('5678::3')], 3),
        ]


def test_fetch_zone_and_subdomain_ipv6_not_implemented(empty_addrfile,
                                                       data_dir):
    """Test that it's a FatalPublishError when neither fetch_zone_ipv6s nor
    fetch_subdomain_ipv6s are implemented"""
    updater = doubles.MockTwoWayZoneUpdater(
        'test_updater', empty_addrfile, data_dir,
        put_zone_ipv6s_result={'example.com': None,
                               'example.net': None},
        put_subdomain_ipv6s_result={('', 'example.com'): None,
                                    ('foo', 'example.com'): None,
                                    ('foo.bar', 'example.net'): None},
    )
    updater.init_hosts_and_zones(
        "example.com foo.bar.example.net foo.example.com"
    )
    with pytest.raises(FatalPublishError):
        updater.publish_ipv6(ipaddress.IPv6Network('5678::/64'))


def test_init_hosts_and_zones_structured(empty_addrfile, data_dir):
    """Test init_hosts_and_zones with structured input instead of raw string"""
    updater = doubles.MockTwoWayZoneUpdater(
        'test_updater', empty_addrfile, data_dir,
        fetch_zone_ipv6s_result={
            'example.com': [
                ('', ipaddress.IPv6Address('1234::1'), 1),
                ('bar', ipaddress.IPv6Address('1234::4'), 4),
                ('foo', ipaddress.IPv6Address('1234::2'), 2),
            ],
            'bar.example.net': [
                ('foo', ipaddress.IPv6Address('1234::3'), 3),
            ],
        },
        put_zone_ipv6s_result={'example.com': None,
                               'bar.example.net': None},
    )
    updater.init_hosts_and_zones([
        ('example.com', None),
        ('foo.bar.example.net', 'bar.example.net'),
        ('bar.example.com', 'example.com'),
        ('foo.example.com', None),
    ])
    updater.publish_ipv6(ipaddress.IPv6Network('5678::/64'))

    assert updater.fetch_zone_ipv6s_calls == [
        'example.com',
        'bar.example.net',
    ]
    assert updater.fetch_subdomain_ipv6s_calls == []
    assert updater.put_zone_ipv6s_calls == [
        ('example.com', {
            '': ([ipaddress.IPv6Address('5678::1')], 1),
            'bar': ([ipaddress.IPv6Address('5678::4')], 4),
            'foo': ([ipaddress.IPv6Address('5678::2')], 2),
        }),
        ('bar.example.net', {
            'foo': ([ipaddress.IPv6Address('5678::3')], 3),
        }),
    ]
    assert updater.put_subdomain_ipv6s_calls == []


# fqdn, zone, subdomain
subdomain_cases = [
    ('', '', ''),
    ('com', 'com', ''),
    ('com', '', 'com'),
    ('example.com', 'example.com', ''),
    ('example.com', 'com', 'example'),
    ('example.com', '', 'example.com'),
    ('www.example.com', 'www.example.com', ''),
    ('www.example.com', 'example.com', 'www'),
    ('www.example.com', 'com', 'www.example'),
    ('www.example.com', '', 'www.example.com'),
]


@pytest.mark.parametrize(('fqdn', 'zone', 'subdomain'), subdomain_cases)
def test_subdomain_of(fqdn, zone, subdomain):
    assert ruddr.TwoWayZoneUpdater.subdomain_of(fqdn, zone) == subdomain


@pytest.mark.parametrize(('fqdn', 'zone', 'subdomain'), subdomain_cases)
def test_fqdn_of(fqdn, zone, subdomain):
    assert ruddr.TwoWayZoneUpdater.fqdn_of(subdomain, zone) == fqdn



#  Ruddr - Robotic Updater for Dynamic DNS Records
#  Copyright (C) 2023 Dominick C. Pastore
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import threading
from typing import List

import pytest

import doubles
import ruddr.addrfile


class VirtualTimer:
    """Drop-in for :class:`threading.Timer` that uses a virtual clock
    instead of wall time and isn't actually threaded. This allows for
    deterministic behavior: the function runs immediately when
    :meth:`start` or :meth:`advance` is called if the virtual time has
    exceeded the interval time."""

    def __init__(self, interval, function, args=None, kwargs=None):
        super().__init__()
        self._function = function
        self._args = args if args is not None else []
        self._kwargs = kwargs if kwargs is not None else {}
        self._interval = interval
        self._lock = threading.Lock()
        self._complete = False
        self._elapsed = 0.0

        self._started = False
        self.daemon = False

    def cancel(self):
        """Stop the timer if it hasn't finished yet."""
        with self._lock:
            self._elapsed = self._interval
            self._complete = True

    def advance(self, seconds):
        """Advance the virtual clock"""
        with self._lock:
            self._elapsed += seconds
            self._try_run()

    @property
    def remaining(self):
        """Number of seconds remaining, or None if timer is complete"""
        with self._lock:
            if self._complete:
                return None
            return max(self._interval - self._elapsed, 0)

    def _try_run(self):
        if self._complete:
            return
        if self._elapsed < self._interval:
            return
        self._function(*self._args, **self._kwargs)
        self._complete = True

    def start(self):
        with self._lock:
            if self._started:
                raise RuntimeError("Already started")
            self._try_run()


@pytest.fixture
def advance():
    """Patch threading.Timer to give us total control of time"""

    class Advancer:
        def __init__(self):
            self.timers: List[VirtualTimer] = []

        def new_timer(self, *args, **kwargs):
            """Create a new virtual timer under the control of this Advancer"""
            timer = VirtualTimer(*args, **kwargs)
            self.timers.append(timer)
            return timer

        def by_minimum_or(self, seconds: float):
            """Advance virtual time just long enough for at least one timer
            to expire or by the given number of seconds, whichever is less, and
            return the number of seconds leftover"""
            remaining_times = [seconds] + [t.remaining for t in self.timers
                                           if t.remaining is not None]
            to_advance = min(remaining_times)
            # Make copy of self.timers so newly created timers aren't advanced
            for timer in list(self.timers):
                timer.advance(to_advance)
            return seconds - to_advance

        def by_minimum(self):
            """Advance virtual time just long enough for at least one timer
            to expire"""
            self.by_minimum_or(float('inf'))

        def by(self, seconds: float):
            """Advance virtual time by the given number of seconds"""
            while seconds > 0:
                seconds = self.by_minimum_or(seconds)

        def until_done(self):
            """Advance virtual time until all timers have elapsed"""
            while any(t.remaining is not None for t in self.timers):
                self.by_minimum()

        def cancel_all(self):
            """Cancel all timers remaining"""
            for timer in self.timers:
                if timer.remaining is not None:
                    timer.cancel()

        def count_running(self):
            """Count the number of timers that have not completed"""
            return sum(t.remaining is not None for t in self.timers)

    advancer = Advancer()

    orig_timer = threading.Timer
    threading.Timer = advancer.new_timer

    yield advancer

    threading.Timer = orig_timer


@pytest.fixture
def notifier_factory():
    """Fixture creating a factory for fake :class:`~ruddr.Notifier`"""
    class NotifierFactory:
        def __init__(self):
            self._count = 0

        def __call__(self, **kwargs):
            self._count += 1
            config = kwargs
            return doubles.FakeNotifier(f'fake_notifier_{self._count}', config)
    return NotifierFactory()


@pytest.fixture
def updater_factory():
    """Fixture creating a factory for mock :class:`~ruddr.Updater`"""
    class UpdaterFactory:
        def __init__(self):
            self._count = 0

        def __call__(self, **kwargs):
            self._count += 1
            return doubles.MockBaseUpdater(f'mock_updater_{self._count}',
                                           **kwargs)
    return UpdaterFactory()


@pytest.fixture
def mock_updater(updater_factory):
    """Fixture creating a mock :class:`~ruddr.Updater` that keeps a list of
    IP addresses it receives"""
    return updater_factory()


@pytest.fixture
def empty_addrfile(tmp_path):
    """Fixture creating an empty :class:`~ruddr.Addrfile`"""
    return ruddr.addrfile.Addrfile(tmp_path / 'addrfile')


@pytest.fixture(scope='session')
def data_dir(tmp_path_factory):
    return str(tmp_path_factory.mktemp('data'))



#  Ruddr - Robotic Updater for Dynamic DNS Records
#  Copyright (C) 2023 Dominick C. Pastore
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""DDNS Manager: Initializes notifiers and updaters and manages the addrfile"""

import importlib
import logging
import os.path
import sys
from typing import Optional, Any, Union, Dict, Tuple, cast

if sys.version_info < (3, 10):
    from importlib_metadata import entry_points
else:
    from importlib.metadata import entry_points

from . import Addrfile
from . import configuration
from .exceptions import NotifierSetupError, ConfigError
from . import notifiers
from . import updaters


class DDNSManager:
    """Manages the rest of the Ruddr system. Creates notifiers and updaters and
    manages the addrfile.

    :param config: A :class:`~ruddr.Config` with the configuration to use

    :raises ConfigError: if configuration is not valid
    """

    def __init__(self, config: configuration.Config):
        self.log = logging.getLogger('ruddr')

        try:
            config.finalize(validate_notifier_type, validate_updater_type)
        except ConfigError as e:
            self.log.critical("Config error: %s", e)
            raise
        self.config = config

        #: Addrfile manager
        addrfile_name = os.path.join(self.config.main['datadir'], 'addrfile')
        self.addrfile = Addrfile(addrfile_name)

        # Creates self.notifiers and self.updaters as dicts
        self.notifiers: Dict[
            str, notifiers.BaseNotifier
        ] = self._create_notifiers()
        self.updaters: Dict[
            str, updaters.BaseUpdater
        ] = self._create_updaters()

        self._discard_unused_notifiers()

    def _create_notifiers(self) -> Dict[str, notifiers.BaseNotifier]:
        """Initialize the notifiers. Assumes the notifiers have been previously
        imported by :func:`validate_notifier_type`."""
        notifiers_dict: Dict[str, notifiers.BaseNotifier] = dict()
        for name, config in self.config.notifiers.items():
            module = config.get('module')
            notifier_type = config['type']

            if module is None:
                notifier_class = notifiers.notifiers[notifier_type]
            else:
                notifier_class = notifiers.notifiers[(module, notifier_type)]

            notifier = notifier_class(name, config)
            notifiers_dict[name] = notifier
        return notifiers_dict

    def _create_updaters(self) -> Dict[str, updaters.BaseUpdater]:
        """Initialize the updaters. Assumes the updaters have been previously
        imported by :func:`validate_updater_type`."""
        updater_dict: Dict[str, updaters.BaseUpdater] = dict()
        for name, config in self.config.updaters.items():
            module = config.get('module')
            updater_type = config['type']

            if module is None:
                updater_class = updaters.updaters[updater_type]
            else:
                updater_class = updaters.updaters[(module, updater_type)]

            updater = updater_class(name, self.addrfile, config)
            do_ipv4, do_ipv6 = self._attach_updater_notifier(updater, config)
            updater.initial_update(do_ipv4, do_ipv6)
            updater_dict[name] = updater
        return updater_dict

    def _attach_updater_notifier(
        self, updater, config
    ) -> Tuple[bool, bool]:
        """Attach the given :class:`~ruddr.Updater` to its notifier(s).
        Assumes config is valid; that is, notifiers all exist.

        :param updater: The :class:`~ruddr.Updater` to be attached
        :param config: That :class:`~ruddr.Updater`'s config

        :return: 2-tuple (ipv4_attached, ipv6_attached)
        """
        ipv4_notifier_name = config.get('notifier4')
        ipv6_notifier_name = config.get('notifier6')
        ipv4_attached = False
        ipv6_attached = False

        if ipv4_notifier_name is not None:
            ipv4_notifier = self.notifiers[ipv4_notifier_name]
            ipv4_attached = ipv4_notifier.attach_ipv4_updater(
                updater.update_ipv4
            )
        if ipv6_notifier_name is not None:
            ipv6_notifier = self.notifiers[ipv6_notifier_name]
            ipv6_attached = ipv6_notifier.attach_ipv6_updater(
                updater.update_ipv6
            )
        return ipv4_attached, ipv6_attached

    def _discard_unused_notifiers(self):
        """Remove notifiers that are not attached to an updater"""
        for name in list(self.notifiers.keys()):
            if not (self.notifiers[name].want_ipv4() or
                    self.notifiers[name].want_ipv6()):
                self.log.warning("Notifier %s not attached to any updater",
                                 name)
                del self.notifiers[name]

    def start(self):
        """Start running all notifiers. Returns after they start. The notifiers
        will continue running in background threads.

        :raises NotifierSetupError: if a notifier fails to start
        """
        self.log.info("Starting all notifiers...")

        for name, notifier in self.notifiers.items():
            try:
                notifier.start()
            except NotifierSetupError:
                self.log.critical("Notifier %s failed to start. Stopping all "
                                  "notifiers.", name)
                for running_notifier in self.notifiers.values():
                    running_notifier.stop()
                raise

        self.log.info("All notifiers started.")

    def do_notify(self):
        """Do an on-demand notify from all notifiers.

        Not all notifiers will support this, but most will.

        Does not raise any exceptions.
        """
        self.log.info("Checking once for all notifiers...")
        for notifier in self.notifiers.values():
            notifier.do_notify()
        self.log.info("Check for all notifiers complete.")

    def stop(self):
        """Stop all running notifiers gracefully. This will allow Python to
        exit naturally.

        Does not raise any exceptions, even if not yet started.
        """
        self.log.info("Stopping all notifiers...")
        for notifier in self.notifiers.values():
            notifier.stop()
        self.log.info("All notifiers stopped.")


def validate_notifier_type(module: Optional[str], type_: str) -> bool:
    """Check if a notifier type exists, importing it for :class:`DDNSManager`
    if it is not one of the built-in notifiers that comes with Ruddr

    :param module: ``None`` for built-in notifiers. Otherwise, the module the
                   notifier can be imported from.
    :param type_: The name of a built-in notifier or the class name of a
                  non-built-in notifier.
    :returns: ``True`` if the notifier exists, ``False`` otherwise
    """
    return _validate_updater_or_notifier_type(
        "notifier",
        notifiers.notifiers,
        module,
        type_
    )


def validate_updater_type(module: Optional[str], type_: str) -> bool:
    """Check if an updater type exists, importing it for :class:`DDNSManager`
    if it is not one of the built-in updaters that comes with Ruddr

    :param module: ``None`` for built-in updaters. Otherwise, the module the
                   updater can be imported from.
    :param type_: The name of a built-in updater or the class name of a
                  non-built-in updater.
    :returns: ``True`` if the updater exists, ``False`` otherwise
    """
    return _validate_updater_or_notifier_type(
        "updater",
        updaters.updaters,
        module,
        type_
    )


def _validate_updater_or_notifier_type(
    which: str,
    existing: Dict[Union[str, Tuple[str, str]], Any],
    module: Optional[str],
    type_: str,
) -> bool:
    """Check if an updater or notifier exists and import it

    :param which: ``"updater"`` or ``"notifier"``
    :param existing: The dict of already known updaters or notifiers and their
                     class. Keys are strings for built-in, (module, class) name
                     tuples for non-built-in
    :param module: ``None`` for built-in updaters or notifiers. Otherwise, the
                   module it can be imported from.
    :param type_: The name of a built-in updater or notifier or the class name
                  of a non-built-in one.
    :returns: ``True`` if the updater or notifier exists, ``False`` otherwise
    """
    if module is None:
        # Check if built-in or already imported with an entry point
        if type_ in existing:
            return True
        # Check if a ruddr entry point with this name exists
        discovered = entry_points(group=f"ruddr.{which}")
        try:
            entry_point = discovered[type_]
        except KeyError:
            return False
        existing[type_] = entry_point.load()
        return True

    # Check if already imported non-built-in
    if (module, type_) in existing:
        return True

    # Check if it's importable
    try:
        imported_module = importlib.import_module(module)
    except ImportError:
        return False
    try:
        imported_class = getattr(imported_module, type_)
    except AttributeError:
        return False
    existing[cast(Tuple[str, str], (module, type_))] = imported_class
    return True



#  Ruddr - Robotic Updater for Dynamic DNS Records
#  Copyright (C) 2023 Dominick C. Pastore
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Base class for Ruddr notifiers"""

import ipaddress
import logging
import threading
# Note: We are not using abstractmethod the way it is intended. We are using it
# purely to get Sphinx to mark methods as abstract. Thus, we intentionally do
# NOT use ABCMeta or inherit from ABC.
from abc import abstractmethod
from typing import Callable, Dict, List, Optional

from ruddr.exceptions import NotifyError, ConfigError, NotStartedError


class BaseNotifier:
    """Base class for all Ruddr notifiers. Sets up the logger, sets up some
    useful member variables, and handles attaching to update functions.

    Notifiers should generally not extend this class directly, but should
    extend :class:`Notifier` instead.

    :param name: Name of the notifier (from config section heading)
    :param config: Dict of config options for this notifier

    :raises ConfigError: if the configuration is invalid
    """

    def __init__(self, name: str, config: Dict[str, str]):
        #: Notifier name (from config section heading)
        self.name: str = name

        #: Logger (see standard :mod:`logging` module)
        self.log = logging.getLogger(f'ruddr.notifier.{self.name}')

        # Configuration options for all Notifiers:
        #
        # - skip_ipv4: default false
        #   If true, this notifier never checks or notifies for IPv4 addresses.
        #   Use when there is no IPv4 networking.
        #
        # - skip_ipv6: default false
        #   Like skip_ipv4, but for IPv6
        #
        # - ipv4_required: default true
        #   When true, failure to obtain the current IPv4 address is an error.
        #   When false, lack of IPv4 address is considered normal.
        #
        # - ipv6_required: default false
        #   Like ipv4_required, but for IPv6. Defaults to false since IPv6
        #   support in networks generally isn't universal.
        #
        # - skip_* options override *_required options
        #
        # Configuration suggestions:
        #
        # - Most people *can* get away with using the defaults, but it's best
        #   to be explicit if possible.
        #
        # - If IPv6 is known to be available, setting ipv6_required ensures
        #   that Ruddr will retry when it can't get a current IPv6 address.
        #
        # - If IPv6 is known to NOT be available, setting skip_ipv6 tells
        #   Ruddr not to waste time trying to determine an IPv6 address.

        try:
            self._skip_ipv4: bool = (config.get('skip_ipv4', 'false').lower()
                                     in ('true', 'on', 'yes', '1'))
        except ValueError:
            self.log.critical("'skip_ipv4' must be boolean (true/yes/on/1/"
                              "false/no/off/0)")
            raise ConfigError(f"'skip_ipv4' option for {self.name} must"
                              "be boolean (true/yes/on/1/false/no/off/0)")
        try:
            self._skip_ipv6: bool = (config.get('skip_ipv6', 'false').lower()
                                     in ('true', 'on', 'yes', '1'))
        except ValueError:
            self.log.critical("'skip_ipv6' must be boolean (true/yes/on/1/"
                              "false/no/off/0)")
            raise ConfigError(f"'skip_ipv6' option for {self.name} must"
                              "be boolean (true/yes/on/1/false/no/off/0)")

        if self._skip_ipv4 and self._skip_ipv6:
            self.log.critical("Cannot skip both IPv4 and IPv6")
            raise ConfigError(f"Notifier {self.name} cannot skip both IPv4 "
                              "and IPv6")

        try:
            self._ipv4_required: bool = (
                config.get('ipv4_required',
                           'false' if self._skip_ipv4 else 'true').lower()
                in ('true', 'on', 'yes', '1')
            )
        except ValueError:
            self.log.critical("'ipv4_required' must be boolean (true/yes/on/1/"
                              "false/no/off/0)")
            raise ConfigError(f"'ipv4_required' option for {self.name} must"
                              "be boolean (true/yes/on/1/false/no/off/0)")
        try:
            self._ipv6_required: bool = (config.get('ipv6_required',
                                                    'false').lower()
                                         in ('true', 'on', 'yes', '1'))
        except ValueError:
            self.log.critical("'ipv6_required' must be boolean (true/yes/on/1/"
                              "false/no/off/0)")
            raise ConfigError(f"'ipv6_required' option for {self.name} must"
                              "be boolean (true/yes/on/1/false/no/off/0)")

        if self._skip_ipv4 and self._ipv4_required:
            self.log.critical("Cannot require IPv4 when it is skipped")
            raise ConfigError(f"{self.name} updater cannot require IPv4 when "
                              "it is skipped")

        if self._skip_ipv6 and self._ipv6_required:
            self.log.critical("Cannot require IPv6 when it is skipped")
            raise ConfigError(f"{self.name} updater cannot require IPv6 when "
                              "it is skipped")

        self._ipv4_update_funcs: List[
            Callable[[ipaddress.IPv4Address], None]
        ] = []
        self._ipv6_update_funcs: List[
            Callable[[ipaddress.IPv6Network], None]
        ] = []

        # Ensure multithreaded notifiers are only notifying once at a time (in
        # case updater is not reentrant)
        self._notify_lock: threading.Lock = threading.Lock()

    def attach_ipv4_updater(
        self, update_func: Callable[[ipaddress.IPv4Address], None],
    ) -> bool:
        """Attach an IPv4 update function to this notifier. No effect if IPv4
        is skipped for this notifier.

        :param update_func: A callable that accepts an :class:`IPv4Address` to
                            be called whenever the IPv4 address might have been
                            updated. It is the callee's responsibility to
                            ensure the update eventually happens successfully
                            (e.g. by scheduling a retry in a separate thread if
                            it fails), but this function should not block
                            longer than a single update attempt.
        :raises ConfigError: If config required for IPv4 notifying is missing

        :return: ``True`` if attached, ``False`` if IPv4 is skipped
        """
        if self._skip_ipv4:
            self.log.info("Not attaching updater to notifier %s for skipped "
                          "IPv4", self.name)
            return False
        self.log.debug("Attaching %s (IPv4) to notifier %s",
                       update_func.__name__, self.name)
        if len(self._ipv4_update_funcs) == 0 and not self.ipv4_ready():
            self.log.critical("Cannot use as IPv4 notifier without required "
                              "IPv4 config")
            raise ConfigError("Notifier %s cannot be an IPv4 notifier without "
                              "required IPv4 config" % self.name)
        self._ipv4_update_funcs.append(update_func)
        return True

    def attach_ipv6_updater(
        self, update_func: Callable[[ipaddress.IPv6Network], None]
    ) -> bool:
        """Attach an IPv6 update function to this notifier. No effect if IPv6
        is skipped for this notifier.

        :param update_func: A callable that accepts an :class:`IPv6Network` to
                            be called whenever the IPv6 prefix might have been
                            updated. It is the callee's responsibility to
                            ensure the update eventually happens successfully
                            (e.g. by scheduling a retry in a separate thread if
                            it fails), but this function should not block
                            longer than a single update attempt.
        :raises ConfigError: If config required for IPv6 notifying is missing

        :return: ``True`` if attached, ``False`` if IPv6 is skipped
        """
        if self._skip_ipv6:
            self.log.info("Not attaching updater to notifier %s for skipped "
                          "IPv6", self.name)
            return False
        self.log.debug("Attaching %s (IPv6) to notifier %s",
                       update_func.__name__, self.name)
        if len(self._ipv6_update_funcs) == 0 and not self.ipv6_ready():
            self.log.critical("Cannot use as IPv6 notifier without required "
                              "IPv4 config")
            raise ConfigError("Notifier %s cannot be an IPv6 notifier without "
                              "required IPv6 config" % self.name)
        self._ipv6_update_funcs.append(update_func)
        return True

    def notify_ipv4(self, address: ipaddress.IPv4Address) -> None:
        """Subclasses must call this to notify all the attached IPv4 updaters
        of a (possibly) new IPv4 address.

        Subclasses may, but need not, call this if :meth:`want_ipv4` is false.

        :param address: The (possibly) new IPv4 address
        """
        self.log.debug("Notifier %s notifying attached updaters of IPv4 %s",
                       self.name, address.exploded)
        with self._notify_lock:
            for update_func in self._ipv4_update_funcs:
                update_func(address)

    def notify_ipv6(self, prefix: ipaddress.IPv6Network) -> None:
        """Subclasses must call this to notify all the attached IPv6 updaters
        of a (possibly) new IPv6 prefix.

        Subclasses may, but need not, call this if :meth:`want_ipv6` is false.

        :param prefix: The (possibly) new IPv6 network prefix
        """
        self.log.debug("Notifier %s notifying attached updaters of IPv6 %s",
                       self.name, prefix.compressed)
        with self._notify_lock:
            for update_func in self._ipv6_update_funcs:
                update_func(prefix)

    def want_ipv4(self) -> bool:
        """Subclasses should call this to determine whether to check for
        current IPv4 addresses at all.

        :return: ``True`` if so, ``False`` if not
        """
        # Will be true if no updaters are configured, or they weren't attached
        # because skip_ipv4
        return len(self._ipv4_update_funcs) > 0

    def want_ipv6(self) -> bool:
        """Subclasses should call this to determine whether to check for
        current IPv6 addresses at all.

        :return: ``True`` if so, ``False`` if not
        """
        # Will be true if no updaters are configured, or they weren't attached
        # because skip_ipv6
        return len(self._ipv6_update_funcs) > 0

    def need_ipv4(self) -> bool:
        """Subclasses must call this to determine if a lack of IPv4 addressing
        is an error.

        :return: ``True`` if so, ``False`` if not
        """
        return self.want_ipv4() and self._ipv4_required

    def need_ipv6(self) -> bool:
        """Subclasses must call this to determine if a lack of IPv6 addressing
        is an error.

        :return: ``True`` if so, ``False`` if not
        """
        return self.want_ipv6() and self._ipv6_required

    @abstractmethod
    def ipv4_ready(self) -> bool:
        """Check if all configuration required for IPv4 notifying is present.

        **Subclasses must override if there is any configuration only required
        for IPv4.**

        :return: ``True`` if so, ``False`` if not
        """
        return True

    @abstractmethod
    def ipv6_ready(self) -> bool:
        """Check if all configuration required for IPv6 notifying is present.

        **Subclasses must override if there is any configuration only required
        for IPv6.**

        :return: ``True`` if so, ``False`` if not
        """
        return True

    @abstractmethod
    def do_notify(self) -> None:
        """Check the IP address a single time and notify immediately, if
        possible.

        This is known as an "on-demand notify." Notifiers should support this
        if possible, but it is okay if they do not. In the latter case, this
        should do nothing and not raise any exceptions (except optionally
        :exc:`NotStartedError`).

        **Must be overridden by subclasses.**

        :raises NotStartedError: if the notifier is not started
        """
        raise NotImplementedError

    @abstractmethod
    def start(self) -> None:
        """Begin ongoing IP address notifications. Any setup should be complete
        before this function returns (e.g. opening socket connections, etc.)
        but ongoing notifications should continue in the background, e.g. on
        a separate thread.

        **Must be overridden by subclasses.**

        :raises NotifierSetupError: when there is a nonrecoverable error
                                    preventing notifier startup.
        """
        raise NotImplementedError

    @abstractmethod
    def stop(self) -> None:
        """Halt ongoing IP address notifications. Clean up gracefully and stop
        any non-daemon threads so Python may exit.

        This must not raise any exceptions, even if called before :meth:`start`
        or after :meth:`start` fails.

        **Must be overridden by subclasses.**
        """
        raise NotImplementedError


class Notifier(BaseNotifier):
    """A base class for notifiers. Supports a variety of notifier strategies,
    such as polling, event-based notifying, and hybrids. See the docs on
    :ref:`writing your own notifier <notifier_dev>` for more detail on what
    that means.

    Also handles setting up a logger, setting some useful member variables for
    subclasses, and attaching to update functions.

    :param name: Name of the notifier
    :param config: Dict of config options for this notifier

    :raises ConfigError: if the configuration is invalid
    """

    def __init__(self, name: str, config: Dict[str, str]):
        super().__init__(name, config)

        # See set_check_intervals for info on these
        self._retry_min_interval: int = 300
        self._retry_max_interval: int = 86400
        self._success_interval: int = 0

        self._lock: threading.RLock = threading.RLock()

        # Ensure can't do_notify, check, or stop when not started, or start
        # when already started. Must lock to access.
        self._started: bool = False

        # Used for retries and polling. Must lock to access.
        self._timer: Optional[threading.Timer] = None
        self._seq: int = 0
        self._retries: int = 0

        # For tests to join the thread doing the first check
        self.first_check = None

    def set_check_intervals(self,
                            retry_min_interval: int = 300,
                            retry_max_interval: int = 86400,
                            success_interval: int = 0,
                            config: Optional[Dict[str, str]] = None) -> None:
        """Set the retry intervals for the :meth:`check_once` function, and
        optionally set :meth:`check_once` to run periodically when successful.

        When :meth:`check_once` fails (by raising :exc:`NotifyError`), the next
        invocation will be scheduled using an exponential backoff strategy,
        starting with ``retry_min_interval`` seconds. Subsequent consecutive
        failures will be scheduled using successively longer intervals, until
        reaching the maximum failure interval, ``retry_max_interval`` seconds.

        The ``success_interval`` parameter triggers some additional behavior:

        - If ``success_interval`` is greater than zero, then when
          :meth:`check_once` succeeds, another invocation will be scheduled for
          ``success_interval`` seconds later. This is useful for notifiers that
          check the current address by polling, like the ``web`` notifier.
          Since :meth:`check` runs at notifier startup, that means
          :meth:`setup` and :meth:`teardown` may not have to be implemented at
          all for these polling-style notifiers.

        - If ``success_interval`` is zero, :meth:`check_once` will only be
          scheduled for future invocation for retries.

        If the ``config`` parameter is provided, it will be checked for keys
        ``retry_min_interval``, ``retry_max_interval``, and ``interval``. If
        either of the ``retry_*`` keys are found, their values override the
        parameters passed into this function. If the ``interval`` key is found,
        its value overrides the ``success_interval`` parameter *if and only if*
        the parameter was already nonzero (preventing a configuration mistake
        from changing a non-polling notifier into a polling notifier).

        In practice, that means this function should be called using the
        notifier's default values for ``retry_min_interval``,
        ``retry_max_interval``, and ``success_interval`` (if it needs defaults
        other than the method defaults) and passing in the config to allow
        the user to override them.

        **Subclasses should call this before returning from their
        constructor (if it needs to be called at all).**

        This has no effect if :meth:`check_once` raises
        :exc:`NotImplementedError`.

        :param retry_min_interval: Minimum retry interval
        :param retry_max_interval: Maximum retry interval
        :param success_interval: Normal polling interval, or 0
        :param config: Config dict for this updater

        :raises ConfigError: if the retry intervals are less than 1, the
                             success interval is less than 0, or the values in
                             the config cannot be converted to :class:`int`
        """
        self._retry_min_interval = retry_min_interval
        self._retry_max_interval = retry_max_interval

        if config is not None:
            try:
                self._retry_min_interval = int(config['retry_min_interval'])
            except KeyError:
                pass
            except ValueError:
                self.log.critical("'retry_min_interval' config option must be "
                                  "an integer > 0")
                raise ConfigError("'retry_min_interval' option for "
                                  f"{self.name} notifier must be an "
                                  "integer > 0")

            try:
                self._retry_max_interval = int(config['retry_max_interval'])
            except KeyError:
                pass
            except ValueError:
                self.log.critical("'retry_max_interval' config option must be "
                                  "an integer > 0")
                raise ConfigError("'retry_max_interval' option for "
                                  f"{self.name} notifier must be an "
                                  "integer > 0")

        if self._retry_min_interval <= 0:
            self.log.critical("'retry_min_interval' config option must be an "
                              "integer > 0")
            raise ConfigError(f"'retry_min_interval' option for {self.name} "
                              "notifier must be an integer > 0")

        if self._retry_max_interval <= 0:
            self.log.critical("'retry_max_interval' config option must be an "
                              "integer > 0")
            raise ConfigError(f"'retry_max_interval' option for {self.name} "
                              "notifier must be an integer > 0")

        self._success_interval = success_interval
        if self._success_interval == 0:
            return
        if config is not None:
            try:
                self._success_interval = int(config['interval'])
            except KeyError:
                pass
            except ValueError:
                self.log.critical("'interval' config option must be an "
                                  "integer >= 0")
                raise ConfigError(f"'interval' option for {self.name} "
                                  "notifier must be an integer > 0")

        if self._success_interval <= 0:
            self.log.critical("'interval' config option must be an "
                              "integer >= 0")
            raise ConfigError(f"'interval' option for {self.name} "
                              "notifier must be an integer > 0")

    def start(self) -> None:
        with self._lock:
            if self._started:
                self.log.warning("Not starting notifier: Already started")
                return

            try:
                self.setup()
            except NotImplementedError:
                self.log.info("Notifier has no setup")
                self._started = True
            else:
                self.log.info("Notifier is finished with setup")
                self._started = True

        # Do the first check in the background
        def first_check():
            try:
                self.check()
            except NotImplementedError:
                self.log.info("Not doing an immediate check as this notifier "
                              "does not support it")
        # Name is only used for testing purposes
        self.first_check = threading.Thread(target=first_check)
        self.first_check.start()

    def stop(self) -> None:
        self.log.info("Stopping notifier")
        with self._lock:
            if not self._started:
                self.log.info("Not stopping notifier: Already stopped")
                return

            # Stop a scheduled retry or check
            self.log.debug("Canceling pending checks")
            if self._timer is not None:
                self._timer.cancel()

            self.log.info("Notifier is starting teardown")
            try:
                self.teardown()
            except NotImplementedError:
                self.log.info("Notifier has no teardown")
                self._started = False
            else:
                self._started = False

    def do_notify(self) -> None:
        self.log.debug("On-demand notify waiting for lock")
        with self._lock:
            self.log.debug("Doing on-demand notify")
            if not self._started:
                self.log.error("Tried to do_notify when not started")
                raise NotStartedError(f"Notifier {self.name} cannot do_notify "
                                      "when not started")
            try:
                self.check()
            except NotImplementedError:
                self.log.info("Notifier does not support on-demand "
                              "notifications")

    @abstractmethod
    def setup(self) -> None:
        """Do any setup and start ongoing IP address notifications. Setup
        should be complete before this function returns (e.g. opening socket
        connections, etc.) but ongoing notifications should continue in the
        background, e.g. on a separate thread.

        **Should be overridden by subclasses if required.**

        :raises NotifierSetupError: when there is a nonrecoverable error
                                    preventing notifier startup.
        """
        raise NotImplementedError

    @abstractmethod
    def teardown(self) -> None:
        """Halt ongoing IP address notifications, do any teardown, and stop any
        non-daemon threads so Python may exit.

        **Should be overridden by subclasses if required.**

        When this is called, there will be no pending invocations of
        :meth:`check_once`, and it's guaranteed that :meth:`setup` is complete.
        Apart from that, it is up to the implementation to ensure that
        inconvenient timing won't break any operations happening in background
        threads (e.g. that were started by :meth:`setup`).

        This must not raise any exceptions (other than
        :exc:`NotImplementedError` if not implemented).
        """
        raise NotImplementedError

    def check(self):
        """Check the current IP address and do a notify, retrying on failure.
        If the ``success_interval`` is nonzero (see
        :meth:`set_check_intervals`) and the check was successful, also
        schedule the next check.

        This is called automatically after notifier startup and for on-demand
        notifies. The notifier can also call this itself to trigger additional
        notifies. For example, the ``systemd`` notifier does this when it
        receives a DBus event saying the network status has changed.

        :raises NotImplementedError: if not supported by this notifier
        """
        self.log.debug("Check waiting for lock")
        with self._lock:
            if not self._started:
                self.log.info("Skipping check when notifier not running")
                return

            self._seq += 1
            self._retries = 0
            self.log.debug("(check seq: %d)", self._seq)
            self._check_and_schedule(self._seq)

    def _scheduled_check(self, seq: int):
        """Do a scheduled check (retry or poll), verifying that no new check
        has happened meanwhile"""
        with self._lock:
            if not self._started:
                # Notifier stopped before we could lock. Abort.
                self.log.debug("(invocation for seq %d aborted due to notifier"
                               " stopping)", seq)
            elif self._seq != seq:
                # Another update has happened in the time since this retry was
                # scheduled. Abort.
                self.log.debug("(invocation for seq %d aborted due to new "
                               "check)", seq)
            else:
                self.log.debug("(new invocation for check seq: %d)", seq)
                self._check_and_schedule(seq)

    def _check_and_schedule(self, seq: int):
        """Do a check, scheduling the next one if necessary after. Do not call
        without holding the lock."""
        try:
            self.check_once()
        except NotifyError:
            # Minimum retry interval the first time, doubling each retry after
            retry_delay = self._retry_min_interval * (2 ** self._retries)
            if retry_delay > self._retry_max_interval:
                retry_delay = self._retry_max_interval
            self._retries += 1
            self.log.info("Check failed. Retrying in %d secs. (seq %d)",
                          retry_delay, seq)
            self._timer = threading.Timer(retry_delay, self._scheduled_check,
                                          args=(seq,))
            self._timer.start()
        else:
            if self._success_interval <= 0:
                self.log.debug("(success, seq %d complete), seq")
                return
            self._retries = 0
            self.log.debug("(success for seq %d, next invocation in %d secs)",
                           seq, self._success_interval)
            self._timer = threading.Timer(self._success_interval,
                                          self._scheduled_check, args=(seq,))
            self._timer.start()

    @abstractmethod
    def check_once(self) -> None:
        """Check the current IP address and do a notify, if possible.

        **Should be overridden by subclasses if supported.**

        Some notifiers do not support notifying on demand (for example, they
        get the current address from an event, thus they can only notify when
        such an event happens). For those updaters, this method should raise
        :exc:`NotImplementedError` when called (which is the default behavior
        when not overridden).

        For any notifier that does support obtaining the current IP address on
        demand, this function should do that immediately and notify using
        :meth:`notify_ipv4` and :meth:`notify_ipv6`. Some additional
        guidelines:

        - Ruddr may not need both IPv4 and IPv6 addresses from this notifier.
          Call :meth:`want_ipv4` and :meth:`want_ipv6` to determine if either
          should be skipped.

        - Even if Ruddr wants an address type, it may or may not be an error if
          it cannot be provided. If :meth:`need_ipv4` returns ``True`` and an
          IPv4 address cannot be obtained, raise :exc:`NotifyError` (after
          notifying for the other address type, if necessary). The same goes
          for :meth:`need_ipv6` and availability of an IPv6 prefix.

        This is called at notifier startup, for on-demand notifies, and any
        time the notifier calls :meth:`check` itself.

        :raises NotifyError: if checking the current IP address failed (will
                             trigger a retry after a delay)
        :raises NotImplementedError: if not supported by this notifier
        """
        raise NotImplementedError
